# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Provides a signals-like system for sending and listening for 'events'


Events are kind of like signals, except they may be listened for on a
global scale, rather than connected on a per-object basis like signals
are. This means that ANY object can emit ANY event, and these events may
be listened for by ANY object.

Events should be emitted AFTER the given event has taken place. Often the
most appropriate spot is immediately before a return statement.
"""

from inspect import ismethod
import logging
import re
import threading
import time
import types
import weakref
from gi.repository import GLib

# define this here so the interpreter doesn't complain
EVENT_MANAGER = None

logger = logging.getLogger(__name__)


class Nothing:
    pass


_NONE = Nothing()  # used by event for a safe None replacement

# Assumes that this module was imported on main thread
_UiThread = threading.current_thread()


def log_event(evty, obj, data):
    """
    Sends an event.

    :param evty: the *type* or *name* of the event.
    :type evty: string
    :param obj: the object sending the event.
    :type obj: object
    :param data: some data about the event, None if not required
    :type data: object
    """
    global EVENT_MANAGER
    e = Event(evty, obj, data)
    EVENT_MANAGER.emit(e)


def add_callback(function, evty=None, obj=None, *args, **kwargs):
    """
    Adds a callback to an event

    You should ALWAYS specify one of the two options on what to listen
    for. While not forbidden to listen to all events, doing so will
    cause your callback to be called very frequently, and possibly may
    cause slowness within the player itself.

    :param function: the function to call when the event happens
    :type function: callable
    :param evty: the *type* or *name* of the event to listen for, eg
            `tracks_added`, `cover_changed`. Defaults to any event if
            not specified.
    :type evty: string
    :param obj: the object to listen to events from, e.g. `exaile.collection`
            or `xl.covers.MANAGER`. Defaults to any object if not
            specified.
    :type obj: object
    :param destroy_with: (keyword arg only) If specified, this event will be
                         detached when the specified Gtk widget is destroyed

    Any additional parameters will be passed to the callback.

    :returns: a convenience function that you can call to remove the callback.
    """
    global EVENT_MANAGER
    return EVENT_MANAGER.add_callback(function, evty, obj, args, kwargs)


def add_ui_callback(function, evty=None, obj=None, *args, **kwargs):
    """
    Adds a callback to an event. The callback is guaranteed to
    always be called on the UI thread.

    You should ALWAYS specify one of the two options on what to listen
    for. While not forbidden to listen to all events, doing so will
    cause your callback to be called very frequently, and possibly may
    cause slowness within the player itself.

    :param function: the function to call when the event happens
    :type function: callable
    :param evty: the *type* or *name* of the event to listen for, eg
            `tracks_added`, `cover_changed`. Defaults to any event if
            not specified.
    :type evty: string
    :param obj: the object to listen to events from, e.g. `exaile.collection`
            or `xl.covers.MANAGER`. Defaults to any object if not
            specified.
    :type obj: object
    :param destroy_with: (keyword arg only) If specified, this event will be
                         detached when the specified Gtk widget is destroyed

    Any additional parameters will be passed to the callback.

    :returns: a convenience function that you can call to remove the callback.
    """
    global EVENT_MANAGER
    return EVENT_MANAGER.add_callback(function, evty, obj, args, kwargs, ui=True)


def remove_callback(function, evty=None, obj=None):
    """
    Removes a callback. Can remove both ui and non-ui callbacks.

    The parameters passed should match those that were passed when adding
    the callback
    """
    global EVENT_MANAGER
    EVENT_MANAGER.remove_callback(function, evty, obj)


class Event:
    """
    Represents an Event
    """

    __slots__ = ['type', 'object', 'data']

    def __init__(self, evty, obj, data):
        """
        evty: the 'type' or 'name' for this Event [string]
        obj: the object emitting the Event [object]
        data: some piece of data relevant to the Event [object]
        """
        self.type = evty
        self.object = obj
        self.data = data


class Callback:
    """
    Represents a callback
    """

    __slots__ = ['wfunction', 'time', 'args', 'kwargs']

    def __init__(self, function, time, args, kwargs):
        """
        @param function: the function to call
        @param time: the time this callback was added
        """
        self.wfunction = _getWeakRef(function)
        self.time = time
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return '<Callback %s>' % self.wfunction()


class _WeakMethod:
    """Represent a weak bound method, i.e. a method doesn't keep alive the
    object that it is bound to. It uses WeakRef which, used on its own,
    produces weak methods that are dead on creation, not very useful.
    Typically, you will use the getRef() function instead of using
    this class directly."""

    def __init__(self, method, notifyDead=None):
        """
        The method must be bound. notifyDead will be called when
        object that method is bound to dies.
        """
        assert ismethod(method)
        if method.__self__ is None:
            raise ValueError("We need a bound method!")
        if notifyDead is None:
            self.objRef = weakref.ref(method.__self__)
        else:
            self.objRef = weakref.ref(method.__self__, notifyDead)
        self.fun = method.__func__

    def __call__(self):
        objref = self.objRef()
        if objref is not None:
            return types.MethodType(self.fun, objref)

    def __eq__(self, method2):
        if not isinstance(method2, _WeakMethod):
            return False
        return (
            self.fun is method2.fun
            and self.objRef() is method2.objRef()
            and self.objRef() is not None
        )

    def __hash__(self):
        return hash(self.fun)

    def __repr__(self):
        dead = ''
        if self.objRef() is None:
            dead = '; DEAD'
        obj = '<%s at %s%s>' % (self.__class__, id(self), dead)
        return obj

    def refs(self, weakRef):
        """Return true if we are storing same object referred to by weakRef."""
        return self.objRef == weakRef


def _getWeakRef(obj, notifyDead=None):
    """
    Get a weak reference to obj. If obj is a bound method, a _WeakMethod
    object, that behaves like a WeakRef, is returned, if it is
    anything else a WeakRef is returned. If obj is an unbound method,
    a ValueError will be raised.
    """
    if ismethod(obj):
        createRef = _WeakMethod
    else:
        createRef = weakref.ref

    if notifyDead is None:
        return createRef(obj)
    else:
        return createRef(obj, notifyDead)


class EventManager:
    """
    Manages all Events
    """

    def __init__(self, use_logger=False, logger_filter=None, verbose=False):
        # sacrifice space for speed in emit
        self.all_callbacks = {}
        self.callbacks = {}
        self.ui_callbacks = {}
        self.use_logger = use_logger
        self.use_verbose_logger = verbose
        self.logger_filter = logger_filter

        # RLock is needed so that event callbacks can themselves send
        # synchronous events and add or remove callbacks
        self.lock = threading.RLock()

        self.pending_ui = []
        self.pending_ui_lock = threading.Lock()

    def emit(self, event):
        """
        Emits an Event, calling any registered callbacks.

        event: the Event to emit [Event]
        """

        emit_logmsg = self.use_logger and (
            not self.logger_filter or re.search(self.logger_filter, event.type)
        )
        emit_verbose = emit_logmsg and self.use_verbose_logger

        global _UiThread
        is_ui_thread = threading.current_thread() == _UiThread

        # note: a majority of the calls to emit are made on the
        #       UI thread

        if is_ui_thread:
            self._emit(event, self.all_callbacks, emit_logmsg, emit_verbose)
        else:
            # Don't issue the log message twice
            with self.pending_ui_lock:
                do_emit = not self.pending_ui
                self.pending_ui.append(
                    (event, self.ui_callbacks, emit_logmsg, emit_verbose)
                )

            if do_emit:
                GLib.idle_add(self._emit_pending)
            self._emit(event, self.callbacks, False, emit_verbose)

    def _emit_pending(self):

        with self.pending_ui_lock:
            events = self.pending_ui
            self.pending_ui = []

        for event in events:
            self._emit(*event)

    def _emit(self, event, exc_callbacks, emit_logmsg, emit_verbose):

        # Accumulate in this set to ensure callbacks only get called once
        callbacks = set()

        with self.lock:
            for tcall in [_NONE, event.type]:
                tcb = exc_callbacks.get(tcall)
                if tcb is not None:
                    for ocall in [_NONE, event.object]:
                        ocb = tcb.get(ocall)
                        if ocb is not None:
                            callbacks.update(ocb)

        # However, do not actually call the callbacks from within the lock
        # -> Otherwise non-ui threads could accidentally block the UI if
        #    they decide to run for too long

        for cb in callbacks:
            try:
                fn = cb.wfunction()
                if fn is None:
                    # Remove callbacks that have been garbage collected.. but
                    # really, should be using remove_callback to clean up after
                    # your event handler
                    with self.lock:
                        try:
                            exc_callbacks[event.type][event.object].remove(cb)
                        except (KeyError, ValueError):
                            pass
                else:
                    if emit_verbose:
                        logger.debug(
                            "Attempting to call "
                            "%(function)s in response "
                            "to %(event)s." % {'function': fn, 'event': event.type}
                        )
                    fn.__call__(
                        event.type, event.object, event.data, *cb.args, **cb.kwargs
                    )
                fn = None
            except Exception:
                # something went wrong inside the function we're calling
                logger.exception("Event callback exception caught!")

        if emit_logmsg:
            logger.debug(
                "Sent '%s' event from %r with data %r",
                event.type,
                event.object,
                event.data,
            )

    def emit_async(self, event):
        """
        Same as emit(), but does not block.
        """
        GLib.idle_add(self.emit, event)

    def add_callback(self, function, evty, obj, args, kwargs, ui=False):
        """
        Registers a callback.
        You should always specify at least one of event type or object.

        @param function: The function to call [function]
        @param evty: The 'type' or 'name' of event to listen for. Defaults
            to any. [string]
        @param obj: The object to listen to events from. Defaults
            to any. [string]

        Returns a convenience function that you can call to
        remove the callback.
        """

        if ui:
            all_cbs = [self.ui_callbacks, self.all_callbacks]
        else:
            all_cbs = [self.callbacks, self.all_callbacks]

        destroy_with = kwargs.pop('destroy_with', None)

        if evty is None:
            evty = _NONE
        if obj is None:
            obj = _NONE

        with self.lock:
            cb = Callback(function, time.time(), args, kwargs)

            # add the specified categories if needed.
            for cbs in all_cbs:
                if evty not in cbs:
                    cbs[evty] = weakref.WeakKeyDictionary()
                try:
                    callbacks = cbs[evty][obj]
                except KeyError:
                    callbacks = cbs[evty][obj] = []

                # add the actual callback
                callbacks.append(cb)

        if self.use_logger:
            if (
                not self.logger_filter
                or evty is _NONE
                or re.search(self.logger_filter, evty)
            ):
                logger.debug("Added callback %s for [%s, %s]" % (function, evty, obj))

        if destroy_with is not None:
            destroy_with.connect(
                'destroy', lambda w: self.remove_callback(function, evty, obj)
            )

        return lambda: self.remove_callback(function, evty, obj)

    def remove_callback(self, function, evty=None, obj=None):
        """
        Unsets a callback

        The parameters must match those given when the callback was
        registered. (minus any additional args)
        """
        if evty is None:
            evty = _NONE
        if obj is None:
            obj = _NONE

        with self.lock:
            for cbs in [self.callbacks, self.all_callbacks, self.ui_callbacks]:
                remove = []
                try:
                    callbacks = cbs[evty][obj]
                    for cb in callbacks:
                        if cb.wfunction() == function:
                            remove.append(cb)
                except KeyError:
                    continue
                except TypeError:
                    continue

                for cb in remove:
                    callbacks.remove(cb)

                if len(callbacks) == 0:
                    del cbs[evty][obj]
                    if len(cbs[evty]) == 0:
                        del cbs[evty]

        if self.use_logger:
            if (
                not self.logger_filter
                or evty is _NONE
                or re.search(self.logger_filter, evty)
            ):
                logger.debug("Removed callback %s for [%s, %s]" % (function, evty, obj))


EVENT_MANAGER = EventManager()

# vim: et sts=4 sw=4



# Copyright (C) 2012 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


from collections import namedtuple
import logging
import sys

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

from xl import event, player, settings as xl_settings
from xl.nls import gettext as _
from xlgui.widgets import info
from xlgui import guiutil

from . import osd_preferences


LOGGER = logging.getLogger(__name__)


Point = namedtuple('Point', 'x y')


def do_assert(is_bool):
    """
    Simulates the `assert` statement
    """
    if not is_bool:
        raise AssertionError()


def _sanitize_window_geometry(
    window, current_allocation, padding, width_fill, height_fill
):
    """
    Sanitizes (x-offset, y-offset, width, height) of the given window,
    to make the window show on the screen.

    :param width_fill, height_fill: specifies the maximum width or height
        of a monitor to fill. 1.0 means "fill the whole monitor"
    :param padding: specifies the padding (from workarea border) to leave empty
    """
    work_area = guiutil.get_workarea_dimensions(window)
    cural = current_allocation
    newal = Gdk.Rectangle()

    newal.x = max(padding, cural.x)
    newal.y = max(padding, cural.y)

    newal.width = min(work_area.width // width_fill, cural.width)
    newal.height = min(work_area.height // height_fill, cural.height)

    newal.x = min(newal.x, work_area.x + work_area.width - newal.width - padding)
    newal.y = min(newal.y, work_area.y + work_area.height - newal.height - padding)

    if newal == cural:
        return None

    if cural.x != newal.x or cural.y != newal.y:
        if cural.width != newal.width or cural.height != newal.height:
            window.get_window().move_resize(newal.x, newal.y, newal.width, newal.height)
        else:
            window.move(newal.x, newal.y)
    else:
        if cural.width != newal.width or cural.height != newal.height:
            window.resize(newal.width, newal.height)
    return newal


class OSDPlugin:
    """
    The plugin for showing an On-Screen Display.
    This object holds all the stuff which may live longer than the window.
    Please note that the window has to be destroyed during plugin runtime,
    see the OSDWindow docstring below for details.
    """

    __window = None
    __css_provider = None
    __options = None

    def enable(self, _exaile):
        """
        Enables the on screen display plugin
        """
        do_assert(self.__window is None)

        # Note: Moving windows will not work on Wayland by design, because Wayland does not know
        # absolute window positioning. Gtk.Window.move() does not work there.
        # See https://blog.gtk.org/2016/07/15/future-of-relative-window-positioning/
        # and https://lists.freedesktop.org/archives/wayland-devel/2015-September/024464.html
        if guiutil.platform_is_wayland():
            raise EnvironmentError("This plugin does not work on Wayland backend.")

        # Cached option values
        self.__options = {
            'background': None,
            'display_duration': None,
            'border_radius': None,
            'use_alpha': False,
        }
        self.__css_provider = Gtk.CssProvider()
        if sys.platform.startswith("win32"):
            # Setting opacity on Windows crashes with segfault,
            # see https://bugzilla.gnome.org/show_bug.cgi?id=674449
            self.__options['use_alpha'] = False
            LOGGER.warning(
                "OSD: Disabling alpha channel because it is not supported on Windows."
            )
        else:
            self.__options['use_alpha'] = True

    def teardown(self, _exaile):
        """
        Shuts down the on screen display plugin
        """
        do_assert(self.__window is not None)
        self.__window.destroy_osd()
        event.remove_callback(self.__on_option_set, 'plugin_osd_option_set')
        event.remove_callback(self.__on_playback_track_start, 'playback_track_start')
        event.remove_callback(self.__on_playback_toggle_pause, 'playback_toggle_pause')
        event.remove_callback(self.__on_playback_player_end, 'playback_player_end')
        event.remove_callback(self.__on_playback_error, 'playback_error')
        self.__window = None

    def disable(self, _exaile):
        """
        Disables the on screen display plugin
        """
        self.teardown(_exaile)

    def on_gui_loaded(self):
        """
        Called when Exaile mostly finished loading
        """
        do_assert(self.__window is None)
        event.add_callback(self.__on_option_set, 'plugin_osd_option_set')
        self.__prepare_osd(False)
        # TODO: OSD looks ugly with CSS not applied on first show. Why is that?

        event.add_callback(self.__on_playback_track_start, 'playback_track_start')
        event.add_callback(self.__on_playback_toggle_pause, 'playback_toggle_pause')
        event.add_callback(self.__on_playback_player_end, 'playback_player_end')
        event.add_callback(self.__on_playback_error, 'playback_error')

    def get_preferences_pane(self):
        """
        Called when the user wants to see the preferences pane for this plugin
        """
        osd_preferences.OSDPLUGIN = self
        return osd_preferences

    def make_osd_editable(self, be_editable):
        """
        Rebuilds the OSD to make it movable and resizable
        """
        do_assert(self.__window is not None)
        self.__window.destroy_osd()
        self.__window = None
        self.__prepare_osd(be_editable)
        self.__window.show_for_a_while()

    def __prepare_osd(self, be_editable):
        do_assert(self.__window is None)
        self.__window = OSDWindow(self.__css_provider, self.__options, be_editable)
        # Trigger initial setup through options.
        for option in (
            'format',
            'background',
            'display_duration',
            'show_progress',
            'position',
            'width',
            'height',
            'border_radius',
        ):
            self.__on_option_set(
                'plugin_osd_option_set',
                xl_settings,
                'plugin/osd/{option}'.format(option=option),
            )
        self.__window.restore_geometry_and_show()

    def __on_option_set(self, _event, settings, option):
        """
        Updates appearance on setting change
        """
        if option == 'plugin/osd/format':
            self.__window.info_area.set_info_format(
                settings.get_option(option, osd_preferences.FormatPreference.default)
            )
        elif option == 'plugin/osd/background':
            if not self.__options['background']:
                self.__options['background'] = Gdk.RGBA()
            rgba = self.__options['background']
            rgba.parse(
                settings.get_option(
                    option, osd_preferences.BackgroundPreference.default
                )
            )
            if self.__options['use_alpha'] is True:
                if rgba.alpha > 0.995:
                    # Bug: We need to set opacity to some value < 1 here
                    # otherwise both corners and fade out transition will look ugly
                    rgba.alpha = 0.99
                    settings.set_option(option, rgba.to_string())
            else:
                if rgba.alpha < 1:
                    rgba.to_color()
                    settings.set_option(option, rgba.to_string())
            GLib.idle_add(self.__update_css_provider)
        elif option == 'plugin/osd/border_radius':
            value = settings.get_option(
                option, osd_preferences.BorderRadiusPreference.default
            )
            self.__window.set_border_width(max(6, value // 2))
            self.__options['border_radius'] = value
            GLib.idle_add(self.__update_css_provider)
            self.__window.emit('size-allocate', self.__window.get_allocation())
        elif option == 'plugin/osd/display_duration':
            self.__options['display_duration'] = int(
                settings.get_option(
                    option, osd_preferences.DisplayDurationPreference.default
                )
            )
        elif option == 'plugin/osd/show_progress':
            self.__window.info_area.set_display_progress(
                settings.get_option(
                    option, osd_preferences.ShowProgressPreference.default
                )
            )
        elif option == 'plugin/osd/position':
            position = Point._make(settings.get_option(option, [20, 20]))
            self.__window.geometry['x'] = position.x
            self.__window.geometry['y'] = position.y
        elif option == 'plugin/osd/width':
            width = settings.get_option(option, 300)
            self.__window.geometry['width'] = width
        elif option == 'plugin/osd/height':
            height = settings.get_option(option, 120)
            self.__window.geometry['height'] = height

    def __update_css_provider(self):
        bgcolor = self.__options['background']
        radius = self.__options['border_radius']
        if bgcolor is None or radius is None:
            return  # seems like we are in early initialization state
        if self.__options['use_alpha'] is True:
            color_str = guiutil.css_from_rgba(bgcolor)
        else:
            color_str = guiutil.css_from_rgba_without_alpha(bgcolor)
        data_str = "window { background-color: %s; border-radius: %spx; }" % (
            color_str,
            str(radius),
        )
        self.__css_provider.load_from_data(data_str.encode('utf-8'))
        return False

    def __on_playback_track_start(self, _event, _player, _track):
        self.__window.show_for_a_while()

    def __on_playback_toggle_pause(self, _event, _player, _track):
        self.__window.show_for_a_while()

    def __on_playback_player_end(self, _event, _player, track):
        if track is None:
            self.__window.hide_immediately()
        else:
            self.__window.show_for_a_while()

    def __on_playback_error(self, _event, _player, _message):
        # TODO: show error instead?
        self.__window.hide_immediately()


plugin_class = OSDPlugin


class OSDWindow(Gtk.Window):
    """
    A popup window showing information of the currently playing track

    Due to the way, the Gtk+ API and some of the many different window managers work,
    the OSD cannot be resizable and movable and have no keyboard focus nor decorations
    at the same time.
    Additionally, in some cases the Gtk+ API specifies that functions may not
    be called after the window has been realized (). In some other cases, Gtk+ API does
    not guarantee that a function works after Gtk.Window.show() is called
    (e.g. Gtk.Window.set_decorated(), set_deletable(), set_titlebar()).
    For these reasons, we need to destroy and rebuild the OSD when we want it to be
    resizable and movable by simple drag operations.

    Another related bug report:
    https://bugzilla.gnome.org/show_bug.cgi?id=782117:
         If a window was initially shown undecorated and set_decorated(True) is called,
         titlebar is drawn inside the window
    """

    __hide_id = None
    __fadeout_id = None
    __autohide = True
    __options = None
    geometry = dict(x=20, y=20, width=300, height=120)  # the default

    def __init__(self, css_provider, options, allow_resize_move):
        """
        Initializes the OSD Window.
        Important: Do not call this constructor before Exaile finished loading,
            otherwise the internal TrackInfoPane will re-render label and icon on each
            `track_tags_changed` event, which causes unnecessary CPU load and delays startup.

        Apply the options after this object was initialized.
        """
        Gtk.Window.__init__(self, type=Gtk.WindowType.TOPLEVEL)
        self.__options = options

        self.set_title('Exaile OSD')
        self.set_keep_above(True)
        self.stick()

        # the next two options don't work on GNOME/Wayland due to a bug
        # between Gtk+ and gnome-shell:
        # https://bugzilla.gnome.org/show_bug.cgi?id=771329
        # there is no API guaranty that they work on other platforms.
        self.set_skip_pager_hint(True)
        self.set_skip_taskbar_hint(True)
        # There is no API guaranty that set_deletable() will work
        self.set_deletable(False)
        self.connect('delete-event', lambda _widget, _event: self.hide_immediately)

        self.connect('screen-changed', self.__on_screen_changed)

        style_context = self.get_style_context()
        style_context.add_provider(
            css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        # Init child widgets
        self.info_area = info.TrackInfoPane(player.PLAYER)
        self.info_area.set_default_text(_('No track played yet'))
        # enable updating OSD contents
        # this is very expensive if done during Exaile startup!
        self.info_area.set_auto_update(True)
        self.info_area.cover.set_property('visible', True)
        # If we don't do this, the label text will be selected if the user
        # pressed the mouse button while the OSD is shown for the first time.
        self.info_area.info_label.set_selectable(False)
        self.info_area.show_all()
        self.add(self.info_area)

        self.__setup_resize_move_related_stuff(allow_resize_move)

        # callbacks needed to show the OSD long enough:
        self.add_events(Gdk.EventMask.LEAVE_NOTIFY_MASK)
        self.connect('leave-notify-event', self.__on_leave_notify_event)

        # Also, maximize, minimize, etc. might happen and we want to undo that
        self.add_events(Gdk.EventMask.STRUCTURE_MASK)
        self.connect('window-state-event', self.__on_window_state_event)

        # Needed to acquire size
        self.info_area.set_display_progress(True)

        # set up the window visual
        self.__on_screen_changed(self, None)

    def __setup_resize_move_related_stuff(self, allow_resize_move):
        # Without decorations, the window cannot be resized on some desktops
        # this especially effects GNOME/Wayland and is probably caused by
        # missing client-side decorations (CSD). This code might break when
        # using client side decorations. In this case, we probably shoud hide
        # the titlebar instead of removing the decorations.
        # Removing decorations is ignored on some platforms to enable
        # the resize grid.
        self.set_decorated(False)
        self.set_resizable(allow_resize_move)

        if allow_resize_move:
            self.set_type_hint(Gdk.WindowTypeHint.NORMAL)
            self.set_title(_("Move or resize OSD"))
            # This is often ignored, but we could try:
            self.connect(
                'realize',
                lambda _widget: self.get_window().set_decorations(
                    Gdk.WMDecoration.RESIZEH | Gdk.WMDecoration.TITLE
                ),
            )
        else:
            # On X11 (at least XWayland), this will make the window be not movable,
            # but makes sure the user can still type.
            self.set_type_hint(Gdk.WindowTypeHint.NOTIFICATION)

        # We do not want to disturb user keyboard input
        self.set_accept_focus(allow_resize_move)

        self.__autohide = not allow_resize_move

    def __on_window_state_event(self, _widget, win_state):
        illegal_states = (
            Gdk.WindowState.FULLSCREEN
            | Gdk.WindowState.ICONIFIED
            | Gdk.WindowState.TILED
            | Gdk.WindowState.MAXIMIZED
            | Gdk.WindowState.BELOW
        )
        if (
            win_state.changed_mask & illegal_states
            and win_state.new_window_state & illegal_states
        ):
            # Just returning Gdk.EVENT_STOP doesn't stop the window manager
            # from changing window state.
            # TODO: This often does not work at all.
            GLib.idle_add(self.restore_geometry_and_show)
            return Gdk.EVENT_STOP
        else:
            return Gdk.EVENT_PROPAGATE

    def destroy_osd(self):
        """
        Cleanups
        """
        # Getting the position can only work on a window being shown on screen
        # This is no problem since the window will be shown permanently during configuration.
        if self.is_visible():
            # X11: Position is off by OSD, because it is fetched with OSD but set without OSD
            #    There is no simple way to fix this because we don't know the OSD geometry.
            # Wayland: Position does not work at all.
            xl_settings.set_option('plugin/osd/position', list(self.get_position()))

            # Don't use Gdk.Window.get_width() here, it may include client-side window decorations!
            # This is not guaranteed to work according to Gtk.Window docs, but it works for me.
            width, height = self.get_size()
            xl_settings.set_option('plugin/osd/width', width)
            xl_settings.set_option('plugin/osd/height', height)
        self.hide_immediately()
        if self.__fadeout_id:
            GLib.source_remove(self.__fadeout_id)
            self.__fadeout_id = None
        if self.__hide_id:
            GLib.source_remove(self.__hide_id)
            self.__hide_id = None
        Gtk.Window.destroy(self)

    def __start_fadeout(self):
        """
        Starts fadeout of the window.
        Hides the window it immediately if fadeout is disabled
        """
        self.__hide_id = None

        gdk_display = self.get_window().get_display()
        # Keep showing the OSD in case the pointer is still over the OSD
        if (
            Gtk.get_major_version() > 3
            or Gtk.get_major_version() == 3
            and Gtk.get_minor_version() >= 20
        ):
            gdk_seat = gdk_display.get_default_seat()
            gdk_device = gdk_seat.get_pointer()
        else:
            gdk_device_manager = gdk_display.get_device_manager()
            gdk_device = gdk_device_manager.get_client_pointer()
        window, _posx, _posy = gdk_device.get_window_at_position()
        if window and window is self.get_window():
            self.show_for_a_while()
            return

        if self.__options['use_alpha'] is True:
            if self.__fadeout_id is None:
                self.__fadeout_id = GLib.timeout_add(30, self.__do_fadeout_step)
        else:
            Gtk.Window.hide(self)
        return False

    def show_for_a_while(self):
        """
        This method makes sure that the OSD is shown. Any previous hiding
        timers or fading transitions will be stopped.
        If hiding is allowed through self.__autohide, a new hiding timer
        will be started.
        """
        # unset potential fadeout process
        if self.__fadeout_id:
            GLib.source_remove(self.__fadeout_id)
            self.__fadeout_id = None
        if Gtk.Widget.get_opacity(self) < 1:
            Gtk.Widget.set_opacity(self, 1)
        # unset potential hide process
        if self.__hide_id:
            do_assert(self.__fadeout_id is None)
            GLib.source_remove(self.__hide_id)
            self.__hide_id = None
        # (re)start hide process
        if self.__autohide:
            self.__hide_id = GLib.timeout_add_seconds(
                self.__options['display_duration'], self.__start_fadeout
            )
        Gtk.Window.present(self)

    def restore_geometry_and_show(self):
        """
        Restores window geometry from options and shows the window afterwards.
        """
        geo = self.geometry
        # automatically resizes to minimum required size
        self.set_default_size(geo['width'], geo['height'])

        self.move(geo['x'], geo['y'])
        self.show_for_a_while()
        # screen size might have changed
        allocation = Gdk.Rectangle()
        allocation.x = geo['x']
        allocation.y = geo['y']
        allocation.width = geo['width']
        allocation.height = geo['height']
        _sanitize_window_geometry(super(Gtk.Window, self), allocation, 10, 0.2, 0.2)

    def set_autohide(self, do_autohide):
        """
        Permanently shows the OSD during configuration.
        This method should only be used from osd_preferences.
        """
        self.__autohide = do_autohide
        if do_autohide:
            do_assert(self.__hide_id is None)
            do_assert(self.__fadeout_id is None)
        GLib.idle_add(self.show_for_a_while)

    def __do_fadeout_step(self):
        """
        Constantly decreases the opacity to fade out the window
        """
        do_assert(self.__hide_id is None)
        if Gtk.Widget.get_opacity(self) > 0.001:
            Gtk.Widget.set_opacity(self, Gtk.Widget.get_opacity(self) - 0.05)
            return True
        else:
            self.__fadeout_id = None
            Gtk.Window.hide(self)
            return False

    def __on_screen_changed(self, _widget, _oldscreen):
        """
        Updates the used colormap
        """
        screen = self.get_screen()
        visual = screen.get_rgba_visual()
        if visual is None:
            # This might happen if there is no X compositor so the X Server
            # does not support transparency
            visual = screen.get_system_visual()
            self.__options['use_alpha'] = False
            LOGGER.warning(
                "OSD: Disabling alpha channel because the Gtk+ "
                "backend does not support it."
            )
        self.set_visual(visual)

    '''
    def __on_size_allocate(self, _widget, _allocation):
        """
            Applies the non-rectangular shape
        """
        # TODO: make this work again
        # Bug in pycairo: cairo_region_* functions are not available before
        # version 1.11.0, see https://bugs.freedesktop.org/show_bug.cgi?id=44336
        # we might want to enable this code below once pycairo is distributed on
        # most Linux distros.
        # cairo_region = cairo.Region.create_rectangle(allocation)
        # as a result, calling
        # self.get_window().shape_combine_region(cairo_region, 0, 0)
        # is impossible. Thus, it is impossible to shape the window.
        # Instead, we have to work around this issue by leaving parts
        # of the window undrawn.

        # leave the old code here for reference:
        width, height = allocation.width, allocation.height
        mask = Gdk.Pixmap(None, width, height, 1)
        context = mask.cairo_create()

        context.set_source_rgb(0, 0, 0)
        context.set_operator(cairo.OPERATOR_CLEAR)
        context.paint()

        radius = self.__options['border_radius']
        inner = (radius, radius, width - radius, height - radius)

        context.set_source_rgb(1, 1, 1)
        context.set_operator(cairo.OPERATOR_SOURCE)
        # Top left corner
        context.arc(inner.x,     inner.y,      radius, 1.0 * pi, 1.5 * pi)
        # Top right corner
        context.arc(inner.width, inner.y,      radius, 1.5 * pi, 2.0 * pi)
        # Bottom right corner
        context.arc(inner.width, inner.height, radius, 0.0 * pi, 0.5 * pi)
        # Bottom left corner
        context.arc(inner.x,     inner.height, radius, 0.5 * pi, 1.0 * pi)
        context.fill()

        self.shape_combine_mask(mask, 0, 0)
    '''

    def __on_leave_notify_event(self, _widget, event_crossing):
        # NotifyType.NONLINEAR means the pointer left the window, not just the widget.
        if event_crossing.detail == Gdk.NotifyType.NONLINEAR:
            self.show_for_a_while()
        return Gdk.EVENT_PROPAGATE

    def hide_immediately(self):
        """
        Immediately hides the OSD and removes all remaining timers or transitions
        """
        if self.__fadeout_id:
            GLib.source_remove(self.__fadeout_id)
            self.__fadeout_id = None
        if self.__hide_id:
            GLib.source_remove(self.__hide_id)
            self.__hide_id = None
        Gtk.Window.hide(self)



# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
    Shared GUI widgets
"""

from collections import namedtuple
from urllib.parse import urlparse

from gi.repository import Gio
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from xl import common, playlist as xl_playlist, trax

from xlgui import icons
from xlgui.guiutil import get_workarea_size


class AttachedWindow(Gtk.Window):
    """
    A window attachable to arbitrary widgets,
    follows the movement of its parent
    """

    __gsignals__ = {'show': 'override'}

    def __init__(self, parent):
        Gtk.Window.__init__(self, Gtk.WindowType.TOPLEVEL)

        self.set_decorated(False)
        self.props.skip_taskbar_hint = True
        self.set_keep_above(True)

        # Only allow resizing
        self.realize()
        self.get_window().set_functions(Gdk.WMFunction.RESIZE)

        self.parent_widget = parent
        self.parent_window_connections = []
        parent.connect('hierarchy-changed', self._on_parent_hierarchy_changed)

    def update_location(self):
        """
        Makes sure the window is
        always fully visible
        """
        workarea = Gdk.Rectangle()
        workarea.x = workarea.y = 0
        workarea.width, workarea.height = get_workarea_size()
        parent_alloc = self.parent_widget.get_allocation()
        toplevel_position = (
            self.parent_widget.get_toplevel().get_window().get_position()
        )
        # Use absolute screen position
        parent_alloc.x += toplevel_position[0]
        parent_alloc.y += toplevel_position[1]

        alloc = self.get_allocation()
        if workarea.width - parent_alloc.x < alloc.width:
            # Parent rightmost
            x = parent_alloc.x + parent_alloc.width - alloc.width
        else:
            # Parent leftmost
            x = parent_alloc.x

        if workarea.height - parent_alloc.y < alloc.height:
            # Parent at bottom
            y = parent_alloc.y - alloc.height
        else:
            # Parent at top
            y = parent_alloc.y + parent_alloc.height

        self.move(x, y)

    def do_show(self):
        """
        Updates the location upon show
        """
        Gtk.Window.do_show(self)
        self.update_location()

    def _on_parent_hierarchy_changed(self, parent_widget, previous_toplevel):
        """(Dis)connect from/to the parent's toplevel window signals"""
        conns = self.parent_window_connections
        for conn in conns:
            previous_toplevel.disconnect(conn)
        conns[:] = ()
        toplevel = parent_widget.get_toplevel()
        if not isinstance(toplevel, Gtk.Window):  # Not anchored
            return
        self.set_transient_for(toplevel)
        conns.append(
            toplevel.connect('configure-event', self._on_parent_window_configure_event)
        )
        conns.append(toplevel.connect('hide', self._on_parent_window_hide))

    def _on_parent_window_configure_event(self, _widget, _event):
        """Update location when parent window is moved"""
        if self.props.visible:
            self.update_location()

    def _on_parent_window_hide(self, _window):
        """Emit the "hide" signal on self when the parent window is hidden.

        If there is a "transient for" relationship between two windows, when
        the parent is hidden, the child is hidden without emitting "hide".
        Here we manually emit it to simplify usage.
        """
        self.emit('hide')


class AutoScrollTreeView(Gtk.TreeView):
    """
    A TreeView which handles autoscrolling upon DnD operations
    """

    def __init__(self):
        Gtk.TreeView.__init__(self)

        self._SCROLL_EDGE_SIZE = 15  # As in gtktreeview.c
        self.__autoscroll_timeout_id = None

        self.connect("drag-motion", self._on_drag_motion)
        self.connect("drag-leave", self._on_drag_leave)

    def _on_drag_motion(self, widget, context, x, y, timestamp):
        """
        Initiates automatic scrolling
        """
        if not self.__autoscroll_timeout_id:
            self.__autoscroll_timeout_id = GLib.timeout_add(
                50, self._on_autoscroll_timeout
            )

    def _on_drag_leave(self, widget, context, timestamp):
        """
        Stops automatic scrolling
        """
        autoscroll_timeout_id = self.__autoscroll_timeout_id

        if autoscroll_timeout_id:
            GLib.source_remove(autoscroll_timeout_id)
            self.__autoscroll_timeout_id = None

    def _on_autoscroll_timeout(self):
        """
        Automatically scrolls during drag operations

        Adapted from gtk_tree_view_vertical_autoscroll() in gtktreeview.c
        """
        _, x, y, _ = self.props.window.get_pointer()
        x, y = self.convert_widget_to_tree_coords(x, y)
        visible_rect = self.get_visible_rect()
        # Calculate offset from the top edge
        offset = y - (
            visible_rect.y + 3 * self._SCROLL_EDGE_SIZE
        )  # 3: Scroll faster upwards

        # Check if we are near the bottom edge instead
        if offset > 0:
            # Calculate offset based on the bottom edge
            offset = y - (
                visible_rect.y + visible_rect.height - 2 * self._SCROLL_EDGE_SIZE
            )

            # Skip if we are not near to top or bottom edge
            if offset < 0:
                return True

        vadjustment = self.get_vadjustment()
        vadjustment.props.value = common.clamp(
            vadjustment.props.value + offset,
            0,
            vadjustment.props.upper - vadjustment.props.page_size,
        )
        self.set_vadjustment(vadjustment)

        return True


class DragTreeView(AutoScrollTreeView):
    """
    A TextView that does easy dragging/selecting/popup menu
    """

    class EventData(
        namedtuple('DragTreeView_EventData', 'event modifier triggers_menu target')
    ):
        """
        Objects that goes inside pending events list
        """

        class Target(
            namedtuple('DragTreeView_EventData_Target', 'path column is_selected')
        ):
            """
            Contains target path info
            """

    targets = [Gtk.TargetEntry.new("text/uri-list", 0, 0)]
    dragged_data = dict()

    def __init__(self, container, receive=True, source=True, drop_pos=None):
        """
        Initializes the tree and sets up the various callbacks
        :param container: The container to place the TreeView into
        :param receive: True if the TreeView should receive drag events
        :param source: True if the TreeView should send drag events
        :param drop_pos: Indicates where a drop operation should occur
                w.r.t. existing entries: 'into', 'between', or None (both).
        """
        AutoScrollTreeView.__init__(self)
        self.container = container
        self.pending_events = []

        if source:
            self.drag_source_set(
                Gdk.ModifierType.BUTTON1_MASK,
                self.targets,
                Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
            )

        if receive:
            self.drop_pos = drop_pos
            self.drag_dest_set(
                Gtk.DestDefaults.ALL,
                self.targets,
                Gdk.DragAction.COPY | Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE,
            )
            self.connect('drag_data_received', self.container.drag_data_received)
            self.connect('drag_data_delete', self.container.drag_data_delete)
        self.receive = receive
        self.drag_context = None
        self.show_cover_drag_icon = True
        self.connect('drag-begin', self.on_drag_begin)
        self.connect('drag-end', self.on_drag_end)
        self.connect('drag-motion', self.on_drag_motion)
        self.connect('button-release-event', self.on_button_release)
        self.connect('button-press-event', self.on_button_press)

        if source:
            self.connect('drag-data-get', self.container.drag_get_data)
            self.drag_source_set_icon_name('gtk-dnd')

    def get_selected_tracks(self):
        """
        Returns the currently selected tracks (stub)
        """
        pass

    def get_target_for(self, event):
        """
        Gets target
        :see: Gtk.TreeView.get_path_at_pos
        :param event: Gdk.Event
        :return: DragTreeView.EventData.Target or None if no target path
        """
        target = self.get_path_at_pos(int(event.x), int(event.y))
        if target:
            return DragTreeView.EventData.Target(
                path=target[0],
                column=target[1],
                is_selected=self.get_selection().path_is_selected(target[0]),
            )

    def set_cursor_at(self, target):
        """
        Sets the cursor at target
        :param target: DragTreeView.EventData.Target
        :return: None
        """
        self.set_cursor(target.path, target.column, False)

    def set_selection_status(self, enabled):
        """
        Change the set selection function
        :param enabled: bool
        :return: None
        """
        self.get_selection().set_select_function(lambda *args: enabled, None)

    def reset_selection_status(self):
        """
        Reset
        :return: None
        """
        self.set_selection_status(True)
        del self.pending_events[:]

    def on_drag_end(self, list, context):
        """
        Called when the dnd is ended
        """
        self.drag_context = None
        self.unset_rows_drag_dest()
        self.drag_dest_set(
            Gtk.DestDefaults.ALL,
            self.targets,
            Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
        )

    def on_drag_begin(self, widget, context):
        """
        Sets the cover of dragged tracks as drag icon
        """
        self.drag_context = context
        Gdk.drag_abort(context, Gtk.get_current_event_time())

        self.reset_selection_status()

        # Load covers
        drag_cover_icon = None
        get_tracks_for_path = getattr(self, 'get_tracks_for_path', None)
        if get_tracks_for_path:
            model, paths = self.get_selection().get_selected_rows()
            drag_cover_icon = icons.MANAGER.get_drag_cover_icon(
                map(get_tracks_for_path, paths)
            )

        if drag_cover_icon is None:
            # Set default icon
            icon_name = (
                'gtk-dnd-multiple'
                if self.get_selection().count_selected_rows() > 1
                else 'gtk-dnd'
            )
            Gtk.drag_set_icon_name(context, icon_name, 0, 0)
        else:
            Gtk.drag_set_icon_pixbuf(context, drag_cover_icon, 0, 0)

    def on_drag_motion(self, treeview, context, x, y, timestamp):
        """
        Called when a row is dragged over this treeview
        """
        if not self.receive:
            return False
        self.enable_model_drag_dest(self.targets, Gdk.DragAction.DEFAULT)
        if self.drop_pos is None:
            return False
        info = treeview.get_dest_row_at_pos(x, y)
        if not info:
            return False
        path, pos = info
        if self.drop_pos == 'into':
            # Only allow dropping into entries.
            if pos == Gtk.TreeViewDropPosition.BEFORE:
                pos = Gtk.TreeViewDropPosition.INTO_OR_BEFORE
            elif pos == Gtk.TreeViewDropPosition.AFTER:
                pos = Gtk.TreeViewDropPosition.INTO_OR_AFTER
        elif self.drop_pos == 'between':
            # Only allow dropping between entries.
            if pos == Gtk.TreeViewDropPosition.INTO_OR_BEFORE:
                pos = Gtk.TreeViewDropPosition.BEFORE
            elif pos == Gtk.TreeViewDropPosition.INTO_OR_AFTER:
                pos = Gtk.TreeViewDropPosition.AFTER
        treeview.set_drag_dest_row(path, pos)
        context.drag_status(context.suggested_action, timestamp)
        return True

    def on_button_press(self, button, event):
        """
        Called when a button is pressed
        """
        # Always grab focus is a workaround to do not loose first click
        self.grab_focus()

        self.reset_selection_status()

        # Only treats 1st button press
        if event.type == Gdk.EventType.BUTTON_PRESS:
            modifier = event.state & Gtk.accelerator_get_default_mod_mask()
            target = self.get_target_for(event)

            if target is None:
                if modifier == 0:
                    # Unselects items if the user press any mouse button on an
                    # empty area of the TreeView and no modifier key is active
                    self.get_selection().unselect_all()

                return True  # Ignore clicks on empty areas

            # Declare
            triggers_menu = event.triggers_context_menu()

            # Disable select function to to do not modify selection
            # Triggering menu will only accept selection
            if target.is_selected and (not modifier or triggers_menu):
                self.set_selection_status(False)

            # If it's not a DnD, it will be treated at button release event
            self.pending_events.append(
                DragTreeView.EventData(event, modifier, triggers_menu, target)
            )

        # Calls `button_press` function on container (if present)
        try:
            button_press_function = self.container.button_press
        except AttributeError:
            return False
        else:
            return button_press_function(button, event)

    def on_button_release(self, button, event):
        """
        Called when a button is released
        Treats the pending events added at button press event

        Handles the popup menu that is displayed when you right click in
        the TreeView list (calls `container.menu` if present)
        """
        self.drag_context = None

        # Get pending event
        try:
            event_data = self.pending_events.pop()
        except IndexError:
            return False

        self.reset_selection_status()

        # Do not set cursor if has a modifier key pressed
        if event_data.modifier == 0 and not (
            event_data.triggers_menu and event_data.target.is_selected
        ):
            self.set_cursor_at(event_data.target)

        if event_data.triggers_menu:
            # Uses menu from container (if present)
            menu = getattr(self.container, 'menu', None)
            if menu:
                menu.popup(event_data.event)
                return True

        return False

    # TODO maybe move this somewhere else? (along with _handle_unknown_drag_data)
    def get_drag_data(self, locs, compile_tracks=True, existing_tracks=[]):
        """
        Handles the locations from drag data

        @param locs: locations we are dealing with (can
            be anything from a file to a folder)
        @param compile_tracks: if true any tracks in the playlists
            that are not found as tracks are added to the list of tracks
        @param existing_tracks: a list of tracks that have already
            been loaded from files (used to skip loading the dragged
            tracks from the filesystem)

        @returns: a 2 tuple in which the first part is a list of tracks
            and the second is a list of playlist (note: any files that are
            in a playlist are not added to the list of tracks, but a track could
            be both in as a found track and part of a playlist)
        """
        # TODO handle if they pass in existing tracks
        trs = []
        playlists = []
        for loc in locs:
            (found_tracks, found_playlist) = self._handle_unknown_drag_data(loc)
            trs.extend(found_tracks)
            playlists.extend(found_playlist)

        if compile_tracks:
            # Add any tracks in the playlist to the master list of tracks
            for playlist in playlists:
                for track in playlist.get_tracks():
                    if track not in trs:
                        trs.append(track)

        return (trs, playlists)

    def _handle_unknown_drag_data(self, loc):
        """
        Handles unknown drag data that has been recieved by
        drag_data_received.  Unknown drag data is classified as
        any loc (location) that is not in the collection of tracks
        (i.e. a new song, or a new playlist)

        @param loc:
            the location of the unknown drag data

        @returns: a 2 tuple in which the first part is a list of tracks
            and the second is a list of playlist
        """
        filetype = None
        info = urlparse(loc)

        # don't use gio to test the filetype if it's a non-local file
        # (otherwise gio will try to connect to every remote url passed in and
        # cause the gui to hang)
        if info.scheme in ('file', ''):
            try:
                filetype = (
                    Gio.File.new_for_uri(loc)
                    .query_info('standard::type', Gio.FileQueryInfoFlags.NONE, None)
                    .get_file_type()
                )
            except GLib.Error:
                filetype = None

        if trax.is_valid_track(loc) or info.scheme not in ('file', ''):
            new_track = trax.Track(loc)
            return ([new_track], [])
        elif xl_playlist.is_valid_playlist(loc):
            # User is dragging a playlist into the playlist list
            # so we add all of the songs in the playlist
            # to the list
            new_playlist = xl_playlist.import_playlist(loc)
            return ([], [new_playlist])
        elif filetype == Gio.FileType.DIRECTORY:
            return (trax.get_tracks_from_uri(loc), [])
        else:  # We don't know what they dropped
            return ([], [])


class ClickableCellRendererPixbuf(Gtk.CellRendererPixbuf):
    """
    Custom :class:`Gtk.CellRendererPixbuf` emitting
    an *clicked* signal upon activation of the pixbuf
    """

    __gsignals__ = {
        'clicked': (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_BOOLEAN,
            (GObject.TYPE_PYOBJECT,),
            GObject.signal_accumulator_true_handled,
        )
    }

    def __init__(self):
        Gtk.CellRendererPixbuf.__init__(self)
        self.props.mode = Gtk.CellRendererMode.ACTIVATABLE

    def do_activate(self, event, widget, path, background_area, cell_area, flags):
        """
        Emits the *clicked* signal
        """
        self.emit('clicked', path)
        return



# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2014-2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GLib
from gi.repository import Gst

import logging
import os
import urllib.parse

from xl import common
from xl import event

from xl.nls import gettext as _

from . import gst_utils
from .dynamic_sink import DynamicAudioSink
from .sink import create_device, priority_boost

from xl.player.engine import ExaileEngine
from xl.player.track_fader import TrackFader
from xl.player.gst import missing_plugin


logger = logging.getLogger(__name__)


class ExaileGstEngine(ExaileEngine):
    """
    Super shiny GStreamer-based engine that does all the things!

    * Audio plugins to modify the output stream
    * gapless playback
    * crossfading (requires gst-plugins-bad)
    * Dynamic audio device switching at runtime

    Notes about crossfading:

    The big change from previous attempts at crossfading is that unlike
    the former unified engine, this tries to depend solely on the playbin
    element to play audio files. The reason for this is that playbin
    is 5000+ lines of battle-hardened C code that handles all of the
    weird edges cases in gstreamer, and we don't wish to duplicate that in
    Exaile if we can avoid it.

    Instead, we can use multiple playbin instances that have duplicate
    output audio devices. This makes crossfading a significantly simpler
    proposition.

    There are two modes for this thing:

    * One is normal/gapless mode (no crossfade), and it uses a normal
      playbin element and controls that directly. The playbin is wrapped
      by the AudioStream object, and it's audio sink is a DynamicAudioSink
      element with the

    * The other is crossfading mode (which requires gst-plugins-bad to be
      installed). Create multiple AudioStream objects, and they have a
      DynamicAudioSink object hooked up to an interaudiosink.

    You can register plugins to modify the output audio via the following
    providers:

    * gst_audio_filter: Multiple instances of this can be created, as they
                        get applied to each stream. It is recommended that
                        plugins inherit from :class:`.ElementBin`
    """

    def __init__(self, name, player, disable_autoswitch):
        ExaileEngine.__init__(self, name, player)
        self.logger = logging.getLogger('%s [%s]' % (__name__, name))

        # Default settings
        self.crossfade_enabled = False
        self.crossfade_duration = 3000

        self.audiosink_device = None
        self.audiosink = None
        self.custom_sink_pipe = None

        # If True, then playback will be stopped if the audio sink changes without
        # the user asking for it
        self.disable_autoswitch = disable_autoswitch

        # This means to fade in when the user plays a track, only enabled
        # when crossfade isn't enabled
        self.user_fade_enabled = False
        self.user_fade_duration = 1000

        # Key: option name; value: attribute on self
        options = {
            '%s/crossfading' % self.name: 'crossfade_enabled',
            '%s/crossfade_duration' % self.name: 'crossfade_duration',
            '%s/audiosink_device' % self.name: 'audiosink_device',
            '%s/audiosink' % self.name: 'audiosink',
            '%s/custom_sink_pipe' % self.name: 'custom_sink_pipe',
            '%s/user_fade_enabled' % self.name: 'user_fade_enabled',
            '%s/user_fade' % self.name: 'user_fade_duration',
        }

        self.settings_unsubscribe = common.subscribe_for_settings(
            self.name, options, self
        )

    #
    # Dynamic properties
    #

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

        if not getattr(self, 'initialized', False):
            return

        if name in ['crossfade_enabled', 'crossfade_duration']:
            self._reconfigure_crossfader()

        if name in ['audiosink_device', 'audiosink', 'custom_sink_pipe']:
            self._reconfigure_sink()

    #
    # API
    #

    def initialize(self):

        object.__setattr__(self, 'initialized', True)

        self.main_stream = AudioStream(self)
        self.other_stream = None
        self.crossfade_out = None

        self.player.engine_load_volume()

        self._reconfigure_crossfader()

    def _reconfigure_crossfader(self):

        self.logger.info("Reconfiguring crossfading")

        cf_duration = None
        if self.crossfade_enabled:
            cf_duration = self.crossfade_duration / 1000.0

            if self.other_stream is None:
                self.other_stream = AudioStream(self)
                self.other_stream.set_user_volume(self.main_stream.get_user_volume())

            self.other_stream.reconfigure_fader(cf_duration, cf_duration)
            self.logger.info("Crossfade: enabled (%sms)", self.crossfade_duration)
        else:
            self.logger.info("Crossfade: disabled")
            if self.other_stream is not None:
                self.other_stream.destroy()

        self.main_stream.reconfigure_fader(cf_duration, cf_duration)

    def _reconfigure_sink(self):

        self.logger.info("Reconfiguring audiosinks")

        self.main_stream.reconfigure_sink()
        if self.other_stream is not None:
            self.other_stream.reconfigure_sink()

    def destroy(self, permanent=True):
        self.main_stream.destroy()

        if self.other_stream is not None:
            self.other_stream.destroy()

        if permanent:
            self.settings_unsubscribe()

        object.__setattr__(self, 'initialized', False)

        self.needs_sink = True
        self.audiosink = None
        self.audiosink_device = None

    #
    # Engine API
    #

    def get_current_track(self):
        return self.main_stream.current_track

    def get_position(self):
        return self.main_stream.get_position()

    def get_state(self):
        state = self.main_stream.get_gst_state()
        if state == Gst.State.PLAYING:
            return 'playing'
        elif state == Gst.State.PAUSED:
            return 'paused'
        else:
            return 'stopped'

    def get_volume(self):
        return self.main_stream.get_user_volume()

    def on_track_stopoffset_changed(self, track):

        for stream in [self.main_stream, self.other_stream]:

            if stream is None or stream.current_track != track:
                continue

            # The fader executes the stop offset, so reconfigure it

            if self.crossfade_enabled:
                stream.reconfigure_fader(
                    self.crossfade_duration, self.crossfade_duration
                )
            else:
                stream.reconfigure_fader(None, None)

    def pause(self):
        self.main_stream.pause()

        if self.other_stream is not None:
            self.other_stream.stop()

    def play(self, track, start_at, paused):
        self._next_track(track, start_at, paused, False, False)

    def seek(self, value):
        return self.main_stream.seek(value)

    def set_volume(self, volume):
        self.main_stream.set_user_volume(volume)
        if self.other_stream is not None:
            self.other_stream.set_user_volume(volume)

    def stop(self):
        if self.other_stream is not None:
            self.other_stream.stop()

        prior_track = self.main_stream.stop(emit_eos=False)
        self.player.engine_notify_track_end(prior_track, True)

    def unpause(self):
        self.main_stream.unpause()

    #
    # Engine private functions
    #

    def _autoadvance_track(self, still_fading=False):

        track = self.player.engine_autoadvance_get_next_track()

        if track:
            play_args = self.player.engine_autoadvance_notify_next(track) + (
                False,
                True,
            )
            self._next_track(*play_args)

        # If still fading, don't stop
        elif not still_fading:
            self.stop()

    @common.idle_add()
    def _eos_func(self, stream):

        if stream == self.main_stream:
            self._autoadvance_track()

    def _error_func(self, stream, msg):
        # Destroy the streams, and create a new one, just in case

        self.player.engine_notify_error(msg)
        self.destroy(permanent=False)
        self.initialize()

    def _next_track(self, track, start_at, paused, already_queued, autoadvance):

        prior_track = self.main_stream.current_track

        # Notify that the track is done
        if prior_track is not None:
            self.player.engine_notify_track_end(prior_track, False)

        if self.crossfade_enabled:
            self.main_stream, self.other_stream = self.other_stream, self.main_stream
            self.main_stream.play(
                track,
                start_at,
                paused,
                already_queued,
                self.crossfade_duration / 1000.0,
                self.crossfade_duration / 1000.0,
            )
            self.other_stream.fader.fade_out_on_play()
        elif self.user_fade_enabled and not autoadvance:
            self.main_stream.play(
                track,
                start_at,
                paused,
                already_queued,
                self.user_fade_duration / 1000.0,
            )
        else:
            self.main_stream.play(track, start_at, paused, already_queued)

        self.player.engine_notify_track_start(track)


class AudioStream:
    """
    An object that can play one or more tracks
    """

    idx = 0

    def __init__(self, engine):

        AudioStream.idx += 1
        self.name = '%s-audiostream-%s' % (engine.name, self.idx)
        self.engine = engine

        self.logger = logging.getLogger(
            '%s [%s-a%s]' % (__name__, engine.name, self.idx)
        )

        #  track being played by this stream
        self.current_track = None
        self.buffered_track = None

        # This exists because if there is a sink error, it doesn't
        # really make sense to recreate the sink -- it'll just fail
        # again. Instead, wait for the user to try to play a track,
        # and maybe the issue has resolved itself (plugged device in?)
        self.needs_sink = True
        self.last_position = 0

        self.audio_filters = gst_utils.ProviderBin(
            'gst_audio_filter', '%s-filters' % self.name
        )

        self.playbin = Gst.ElementFactory.make("playbin", "%s-playbin" % self.name)
        if self.playbin is None:
            raise TypeError("gstreamer 1.x base plugins not installed!")

        gst_utils.disable_video_text(self.playbin)

        self.playbin.connect("about-to-finish", self.on_about_to_finish)

        video = Gst.ElementFactory.make("fakesink", '%s-fakevideo' % self.name)
        video.set_property('sync', True)
        self.playbin.set_property('video-sink', video)

        self.audio_sink = DynamicAudioSink('%s-sink' % self.name)
        self.playbin.set_property('audio-sink', self.audio_sink)

        # Setup the bus
        bus = self.playbin.get_bus()
        bus.add_signal_watch()
        bus.connect('message', self.on_message)

        # priority boost hack if needed
        priority_boost(self.playbin)

        # Pulsesink changes volume behind our back, track it
        self.playbin.connect('notify::volume', self.on_volume_change)

        self.fader = TrackFader(
            self, self.on_fade_out_begin, '%s-fade-%s' % (engine.name, self.idx)
        )

    def destroy(self):

        self.fader.stop()
        self.playbin.set_state(Gst.State.NULL)
        self.playbin.get_bus().remove_signal_watch()

    def reconfigure_sink(self):
        self.needs_sink = False
        sink = create_device(self.engine.name)

        # Works for pulsesink, but not other sinks
        # -> Not a perfect solution, still some audio blip is heard. Unfortunately,
        #    can't do better without direct support from gstreamer
        if self.engine.disable_autoswitch and hasattr(sink.props, 'current_device'):
            self.selected_sink = sink.props.device
            sink.connect('notify::current-device', self._on_sink_change_notify)

        self.audio_sink.reconfigure(sink)

    def _on_sink_change_notify(self, sink, param):
        if self.selected_sink != sink.props.current_device:
            domain = GLib.quark_from_string("g-exaile-error")
            err = GLib.Error.new_literal(domain, "Audio device disconnected", 0)
            self.playbin.get_bus().post(
                Gst.Message.new_error(None, err, "Disconnected")
            )
            self.logger.info("Detected device disconnect, stopping playback")

    def reconfigure_fader(self, fade_in_duration, fade_out_duration):
        if self.get_gst_state() != Gst.State.NULL:
            self.fader.setup_track(
                self.current_track, fade_in_duration, fade_out_duration, is_update=True
            )

    def get_gst_state(self):
        return self.playbin.get_state(timeout=50 * Gst.MSECOND)[1]

    def get_position(self):
        # TODO: This only works when pipeline is prerolled/ready?
        if not self.get_gst_state() == Gst.State.PAUSED:
            res, self.last_position = self.playbin.query_position(Gst.Format.TIME)

            if res is False:
                self.last_position = 0

        return self.last_position

    def get_volume(self):
        return self.playbin.props.volume

    def get_user_volume(self):
        return self.fader.get_user_volume()

    def pause(self):
        # This caches the current last position before pausing
        self.get_position()
        self.playbin.set_state(Gst.State.PAUSED)
        self.fader.pause()

    def play(
        self,
        track,
        start_at,
        paused,
        already_queued,
        fade_in_duration=None,
        fade_out_duration=None,
    ):
        '''fade duration is in seconds'''

        if not already_queued:
            self.stop(emit_eos=False)

            # For the moment, the only safe time to add/remove elements
            # is when the playbin is NULL, so do that here..
            if self.audio_filters.setup_elements():
                self.logger.debug("Applying audio filters")
                self.playbin.props.audio_filter = self.audio_filters
            else:
                self.logger.debug("Not applying audio filters")
                self.playbin.props.audio_filter = None

        if self.needs_sink:
            self.reconfigure_sink()

        self.current_track = track
        self.last_position = 0
        self.buffered_track = None

        uri = track.get_loc_for_io()
        self.logger.info("Playing %s", common.sanitize_url(uri))

        # This is only set for gapless playback
        if not already_queued:
            self.playbin.set_property("uri", uri)
            if urllib.parse.urlsplit(uri)[0] == "cdda":
                self.notify_id = self.playbin.connect(
                    'source-setup', self.on_source_setup, track
                )

        # Start in paused mode if we need to seek
        if paused or start_at is not None:
            self.playbin.set_state(Gst.State.PAUSED)
        elif not already_queued:
            self.playbin.set_state(Gst.State.PLAYING)

        self.fader.setup_track(track, fade_in_duration, fade_out_duration, now=0)

        if start_at is not None:
            self.seek(start_at)
            if not paused:
                self.playbin.set_state(Gst.State.PLAYING)

        if paused:
            self.fader.pause()

    def seek(self, value):
        '''value is in seconds'''

        # TODO: Make sure that we're in a valid seekable state before seeking?

        # wait up to 1s for the state to switch, else this fails
        if (
            self.playbin.get_state(timeout=1000 * Gst.MSECOND)[0]
            != Gst.StateChangeReturn.SUCCESS
        ):
            # TODO: This error message is misleading, when does this ever happen?
            # TODO: if the sink is incorrectly specified, this error happens first.
            # self.engine._error_func(self, "Could not start at specified offset")
            self.logger.warning("Error seeking to specified offset")
            return False

        new_position = int(Gst.SECOND * value)
        seek_event = Gst.Event.new_seek(
            1.0,
            Gst.Format.TIME,
            Gst.SeekFlags.FLUSH,
            Gst.SeekType.SET,
            new_position,
            Gst.SeekType.NONE,
            0,
        )

        self.last_position = new_position
        self.fader.seek(value)

        return self.playbin.send_event(seek_event)

    def set_volume(self, volume):
        # self.logger.debug("Set playbin volume: %.2f", volume)
        # TODO: strange issue where pulse sets the system audio volume
        #       when exaile starts up...
        self.playbin.props.volume = volume

    def set_user_volume(self, volume):
        self.logger.debug("Set user volume: %.2f", volume)
        self.fader.set_user_volume(volume)

    def stop(self, emit_eos=True):
        prior_track = self.current_track
        self.current_track = None
        self.playbin.set_state(Gst.State.NULL)
        self.fader.stop()

        if emit_eos:
            self.engine._eos_func(self)

        return prior_track

    def unpause(self):

        # gstreamer does not buffer paused network streams, so if the user
        # is unpausing a stream, just restart playback
        current = self.current_track
        if not (current.is_local() or current.get_tag_raw('__length')):
            self.playbin.set_state(Gst.State.READY)

        self.playbin.set_state(Gst.State.PLAYING)
        self.fader.unpause()

    #
    # Events
    #

    def on_about_to_finish(self, *args):
        """
        This function exists solely to allow gapless playback for audio
        formats that support it. Setting the URI property of the playbin
        will queue the track for playback immediately after the previous
        track.

        .. note:: This is called from the gstreamer thread
        """

        if self.engine.crossfade_enabled:
            return

        track = self.engine.player.engine_autoadvance_get_next_track(gapless=True)
        if track:
            uri = track.get_loc_for_io()
            self.playbin.set_property('uri', uri)
            self.buffered_track = track

            self.logger.debug(
                "Gapless transition: queuing %s", common.sanitize_url(uri)
            )

    def on_fade_out_begin(self):

        if self.engine.crossfade_enabled:
            self.engine._autoadvance_track(still_fading=True)

    def on_message(self, bus, message):
        """
        This is called on the main thread
        """

        if message.type == Gst.MessageType.BUFFERING:
            percent = message.parse_buffering()
            if not percent < 100:
                self.logger.info('Buffering complete')
            if percent % 5 == 0:
                event.log_event('playback_buffering', self.engine.player, percent)

        elif message.type == Gst.MessageType.TAG:
            """Update track length and optionally metadata from gstreamer's parser.
            Useful for streams and files mutagen doesn't understand."""

            current = self.current_track

            if not current.is_local():
                gst_utils.parse_stream_tags(current, message.parse_tag())

            if current and not current.get_tag_raw('__length'):
                res, raw_duration = self.playbin.query_duration(Gst.Format.TIME)
                if not res:
                    self.logger.error("Couldn't query duration")
                    raw_duration = 0
                duration = float(raw_duration) / Gst.SECOND
                if duration > 0:
                    current.set_tag_raw('__length', duration)

        elif (
            message.type == Gst.MessageType.EOS
            and not self.get_gst_state() == Gst.State.PAUSED
        ):
            self.engine._eos_func(self)

        elif (
            message.type == Gst.MessageType.STREAM_START
            and message.src == self.playbin
            and self.buffered_track is not None
        ):

            # This handles starting the next track during gapless transition
            buffered_track = self.buffered_track
            self.buffered_track = None
            play_args = self.engine.player.engine_autoadvance_notify_next(
                buffered_track
            ) + (True, True)
            self.engine._next_track(*play_args)

        elif message.type == Gst.MessageType.STATE_CHANGED:

            # This idea from quodlibet: pulsesink will not notify us when
            # volume changes if the stream is paused, so do it when the
            # state changes.
            if message.src == self.audio_sink:
                self.playbin.notify("volume")

        elif message.type == Gst.MessageType.ERROR:
            self.__handle_error_message(message)

        elif message.type == Gst.MessageType.ELEMENT:
            if not missing_plugin.handle_message(message, self.engine):
                logger.debug(
                    "Unexpected element-specific GstMessage received from %s: %s",
                    message.src,
                    message,
                )

        elif message.type == Gst.MessageType.WARNING:
            # TODO there might be some useful warnings we ignore for now.
            gerror, debug_text = Gst.Message.parse_warning(message)
            logger.warning(
                "Unhandled GStreamer warning received:\n\tGError: %s\n\tDebug text: %s",
                gerror,
                debug_text,
            )

        else:
            # TODO there might be some useful messages we ignore for now.
            logger.debug(
                "Unhandled GstMessage of type %s received: %s", message.type, message
            )

    def __handle_error_message(self, message):
        # Error handling code is from quodlibet
        gerror, debug_info = message.parse_error()
        message_text = ""
        if gerror:
            message_text = gerror.message.rstrip(".")

        if message_text == "":
            # The most readable part is always the last..
            message_text = debug_info[debug_info.rfind(':') + 1 :]

            # .. unless there's nothing in it.
            if ' ' not in message_text:
                if debug_info.startswith('playsink'):
                    message_text += _(
                        ': Possible audio device error, is it plugged in?'
                    )

        self.logger.error("Playback error: %s", message_text)
        self.logger.debug("- Extra error info: %s", debug_info)

        envname = 'GST_DEBUG_DUMP_DOT_DIR'
        if envname not in os.environ:
            import xl.xdg

            os.environ[envname] = xl.xdg.get_logs_dir()

        Gst.debug_bin_to_dot_file(self.playbin, Gst.DebugGraphDetails.ALL, self.name)
        self.logger.debug(
            "- Pipeline debug info written to file '%s/%s.dot'",
            os.environ[envname],
            self.name,
        )

        self.engine._error_func(self, message_text)

    def on_source_setup(self, playbin, source, track):
        # this is for handling multiple CD devices properly
        device = track.get_loc_for_io().split("#")[-1]
        source.props.device = device
        playbin.disconnect(self.notify_id)

    def on_volume_change(self, e, p):
        real = self.playbin.props.volume
        vol, is_same = self.fader.calculate_user_volume(real)
        if not is_same:
            GLib.idle_add(self.engine.player.engine_notify_user_volume_change, vol)



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import os
from typing import Optional
import urllib.parse

from gi.repository import Gio

from xl.metadata._base import BaseFormat, CoverImage, NotWritable, NotReadable

from xl.metadata import (
    aiff,
    ape,
    asf,
    flac,
    mka,
    mod,
    mp3,
    mp4,
    mpc,
    ogg,
    sid,
    speex,
    tta,
    wav,
    wv,
)

#: dictionary mapping extensions to Format classes.
formats = {
    # fmt: off
    '669'   : mod.ModFormat,
    'aac'   : mp4.MP4Format,
    'ac3'   : None,
    'aif'   : aiff.AIFFFormat,
    'aifc'  : aiff.AIFFFormat,
    'aiff'  : aiff.AIFFFormat,
    'ape'   : ape.MonkeysFormat,
    'amf'   : mod.ModFormat,
    'asf'   : asf.AsfFormat,
    'au'    : wav.WavFormat,
    'dsm'   : mod.ModFormat,
    'far'   : mod.ModFormat,
    'flac'  : flac.FlacFormat,
    'it'    : mod.ModFormat,
    'm4a'   : mp4.MP4Format,
    'med'   : mod.ModFormat,
    'mka'   : mka.MkaFormat,
    'mp2'   : mp3.MP3Format,
    'mp3'   : mp3.MP3Format,
    'mp4'   : mp4.MP4Format,
    'mpc'   : mpc.MpcFormat,
    'mid'   : None,
    'midi'  : None,
    'mod'   : mod.ModFormat,
    'mtm'   : mod.ModFormat,
    'oga'   : ogg.OggFormat,
    'ogg'   : ogg.OggFormat,
    'ogx'   : ogg.OggFormat,
    'okt'   : mod.ModFormat,
    'opus'  : ogg.OggOpusFormat,
    'ra'    : None,
    'ram'   : None,
    's3m'   : mod.ModFormat,
    'sid'   : sid.SidFormat,
    'shn'   : None,
    'snd'   : wav.WavFormat,
    'spc'   : None,
    'spx'   : speex.SpeexFormat,
    'stm'   : mod.ModFormat,
    'tta'   : tta.TTAFormat,
    'ult'   : mod.ModFormat,
    'wav'   : wav.WavFormat,
    'wma'   : asf.AsfFormat,
    'wv'    : wv.WavpackFormat,
    'xm'    : mod.ModFormat,
    # fmt: on
}


# Dummy format class to use with unsupported format instead of directly
# instantiating BaseFormat, which may cause issues with tag mapping
# initialization in subclasses (see issue #539)
class DummyFormat(BaseFormat):
    pass


def get_format(loc: str) -> Optional[BaseFormat]:
    """
    get a Format object appropriate for the file at loc.
    if no suitable object can be found, None is returned.

    :param loc: The location to read from as a Gio URI
        (from Track.get_loc_for_io())
    """
    loc = Gio.File.new_for_uri(loc).get_path()
    if not loc:
        return None

    ext = os.path.splitext(loc)[1]
    ext = ext[1:]  # remove the pesky .
    ext = ext.lower()

    try:
        formatclass = formats[ext]
    except KeyError:
        return None  # not supported

    if formatclass is None:
        formatclass = DummyFormat

    try:
        return formatclass(loc)
    except NotReadable:
        return None


# vim: et sts=4 sw=4



# Copyright (C) 2011 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import time

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

from xl import event, providers, settings

from xl.nls import gettext as _
from xlgui.guiutil import GtkTemplate
import xlgui.main
from xlgui.widgets import menu, dialogs

from . import bpmdetect

autodetect_enabled = bpmdetect.autodetect_supported()

menu_providers = ['track-panel-menu', 'playlist-context-menu']


class BPMCounterPlugin:
    """
    Implements logic for plugin
    """

    # Provider API requirement
    name = 'BPM'
    menuitem = None

    def enable(self, exaile):
        pass

    def on_gui_loaded(self):
        providers.register('mainwindow-info-area-widget', self)

        if autodetect_enabled:
            self.menuitem = menu.simple_menu_item(
                '_bpm',
                ['enqueue'],
                _('Autodetect BPM'),
                callback=self.on_auto_menuitem,
                condition_fn=lambda n, p, c: not c['selection-empty'],
            )

            for p in menu_providers:
                providers.register(p, self.menuitem)

    def disable(self, exaile):
        """
        Called when the plugin is disabled
        """
        providers.unregister('mainwindow-info-area-widget', self)

        if self.menuitem is not None:
            for p in menu_providers:
                providers.unregister(p, self.menuitem)

    def create_widget(self, info_area: xlgui.main.MainWindowTrackInfoPane):
        """
        mainwindow-info-area-widget provider API method
        """
        return BPMWidget(info_area.get_player(), self)

    def on_auto_menuitem(self, menu, display_name, playlist_view, context):
        tracks = context['selected-tracks']
        if len(tracks) > 0:
            if isinstance(playlist_view, Gtk.Widget):
                window = playlist_view.get_toplevel()
            else:
                from xlgui import panel

                if isinstance(playlist_view, panel.Panel):
                    window = playlist_view.parent
                else:
                    window = None
            self.autodetect_bpm(tracks[0], window)

    def autodetect_bpm(self, track, parent_window=None):
        def _on_complete(bpm, err):
            if err is not None:
                dialogs.error(None, err)
            elif track and bpm:
                bpm = int(round(float(bpm)))
                msg = BPMAutodetectResponse(parent_window, bpm, track)
                result = msg.run()
                bpm = msg.get_bpm()
                msg.destroy()
                self._set_bpm(result, bpm, track)

        bpmdetect.detect_bpm(track.get_loc_for_io(), _on_complete)

    def set_bpm(self, track, bpm, parent_window=None):
        '''Make sure we don't accidentally set BPM on things'''

        if track and bpm:
            # bpm at this point can be a string or a number.
            # Turn it into a rounded int.
            bpm = int(round(float(bpm)))

            msg = Gtk.MessageDialog(
                buttons=Gtk.ButtonsType.YES_NO,
                message_type=Gtk.MessageType.QUESTION,
                modal=True,
                text=_('Set BPM of %d on %s?') % (bpm, track.get_tag_display('title')),
                transient_for=parent_window,
            )
            msg.set_default_response(Gtk.ResponseType.NO)
            result = msg.run()
            msg.destroy()
            self._set_bpm(result, bpm, track)

    def _set_bpm(self, result, bpm, track):
        if result == Gtk.ResponseType.YES:
            track.set_tags(bpm=bpm)
            if not track.write_tags():
                dialogs.error(
                    None,
                    "Error writing BPM to %s"
                    % GLib.markup_escape_text(track.get_loc_for_io()),
                )


plugin_class = BPMCounterPlugin


@GtkTemplate('msg.ui', relto=__file__)
class BPMAutodetectResponse(Gtk.Dialog):

    __gtype_name__ = 'BPMAutodetectResponse'

    q_label, r1, r2, r3 = GtkTemplate.Child.widgets(4)

    def __init__(self, parent_window, bpm, track):
        Gtk.Dialog.__init__(self, parent=parent_window)
        self.init_template()

        self.q_label.set_text(self.q_label.get_text() % track.get_tag_display('title'))
        self.r1.set_label(str(int(round(bpm / 2.0))))
        self.r2.set_label(str(bpm))
        self.r3.set_label(str(int(round(bpm * 2.0))))
        self.r2.set_active(True)

    def get_bpm(self):
        for r in [self.r1, self.r2, self.r3]:
            if r.get_active():
                return int(r.get_label())


@GtkTemplate('bpm.ui', relto=__file__)
class BPMWidget(Gtk.Frame):

    __gtype_name__ = 'BPMWidget'

    eventbox, bpm_label, apply_button = GtkTemplate.Child.widgets(3)

    def __init__(self, player, plugin):
        Gtk.Frame.__init__(self, label=_('BPM Counter'))
        self.init_template()

        self.player = player
        self.plugin = plugin
        self.taps = []

        # TODO: Add preferences to adjust these settings..

        # number of seconds to average taps over so it converges faster
        self.tap_history = settings.get_option('plugin/bpm/tap_history', 5.0)

        # if no tap received, then restart
        self.stale_time = settings.get_option('plugin/bpm/stale_period', 2.0)

        # Autodetect plugin

        self.menu = None
        if autodetect_enabled:
            self.menu = menu.Menu(None)

            item = menu.simple_menu_item(
                '_bpm', [], _('Autodetect BPM'), callback=self.on_auto_menuitem
            )
            self.menu.add_item(item)

        # Be notified when a new track is playing
        event.add_ui_callback(
            self.playback_track_start, 'playback_track_start', self.player
        )

        # get the main exaile window, and dock our window next to it if possible

        # trigger start event if exaile is currently playing something
        if self.player.is_playing():
            self.playback_track_start(None, self.player, self.player.current)
        else:
            self.track = None
            self.bpm = None
            self.taps = []
            self.update_ui()

    #
    # Exaile events
    #

    def playback_track_start(self, type, player, track):
        self.track = track
        self.bpm = self.track.get_tag_raw('bpm', True)
        self.taps = []

        self.update_ui(False)

    #
    # UI Events
    #

    @GtkTemplate.Callback
    def on_destroy(self, widget):
        # de-register the exaile events
        event.remove_callback(
            self.playback_track_start, 'playback_track_start', self.player
        )

    @GtkTemplate.Callback
    def on_apply_button_clicked(self, widget):
        self.set_bpm()

    @GtkTemplate.Callback
    def on_eventbox_key_press_event(self, widget, event):

        if event.keyval == Gdk.KEY_Return:
            self.set_bpm()
            return False

        if widget == self.apply_button:
            return False

        if event.keyval == Gdk.KEY_Escape:
            self.taps = []

        self.add_bpm_tap()
        return True

    @GtkTemplate.Callback
    def on_eventbox_button_press_event(self, widget, event):
        if event.type is not Gdk.EventType.BUTTON_PRESS:
            return False  # Ignore double- and triple-click events

        if event.triggers_context_menu():
            if self.menu is not None and self.track is not None:
                self.menu.popup(event)
            return

        self.eventbox.set_state(Gtk.StateType.SELECTED)
        self.eventbox.grab_focus()

        self.add_bpm_tap()
        return True

    @GtkTemplate.Callback
    def on_eventbox_focus_out_event(self, widget, event):
        self.eventbox.set_state(Gtk.StateType.NORMAL)

    def on_auto_menuitem(self, *args):
        if self.track is not None:
            self.plugin.autodetect_bpm(self.track, self.get_toplevel())

    #
    # BPM Logic
    #

    def add_bpm_tap(self):
        '''Recalculates the BPM each time an event occurs'''

        current = time.time()

        if len(self.taps) > 0:
            # reset counter if its stale
            if current - self.taps[-1] > self.stale_time:
                self.taps = []

        self.taps.append(current)
        self.trim_taps()

        if len(self.taps) > 1:
            self.bpm = str(
                int(
                    round(
                        ((len(self.taps) - 1) * 60.0) / (self.taps[-1] - self.taps[0])
                    )
                )
            )
        else:
            self.bpm = None

        self.update_ui()

    def trim_taps(self):
        '''Remove old taps so the BPM value converges faster'''
        while len(self.taps) != 0 and self.taps[-1] - self.taps[0] > self.tap_history:
            self.taps.pop(0)

    def set_bpm(self):
        '''Make sure we don't accidentally set BPM on things'''
        self.plugin.set_bpm(self.track, self.bpm, parent_window=self.get_toplevel())
        self.update_ui()

    def update_ui(self, apply_enabled=True):
        '''Updates the current UI display'''

        if self.bpm is None:
            self.bpm_label.set_label(_('Update'))
            self.apply_button.set_sensitive(False)
        else:
            self.bpm_label.set_label(self.bpm)

            if self.track is not None:
                self.apply_button.set_sensitive(apply_enabled)
            else:
                self.apply_button.set_sensitive(False)



# Copyright (C) 2009-2010 Erin Drummond
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GObject
from gi.repository import Gtk

from . import jamtree
from . import jamapi
from . import menu
import os
from xl import common, event, settings, providers, trax as xltrack
from xl.covers import CoverSearchMethod
from xl.nls import gettext as _
from xlgui import icons, panel
from xlgui.widgets.common import DragTreeView

JAMENDO_NOTEBOOK_PAGE = None
COVERS_METHOD = None


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, 'exaile_loaded')
    else:
        _enable(None, exaile, None)


def _enable(eventname, exaile, nothing):
    global JAMENDO_NOTEBOOK_PAGE, COVERS_METHOD

    user_agent = exaile.get_user_agent_string('jamendo')
    jamapi.set_user_agent(user_agent)

    JAMENDO_NOTEBOOK_PAGE = JamendoPanel(exaile.gui.main.window, exaile)
    COVERS_METHOD = JamendoCoverSearch(user_agent)
    providers.register('main-panel', JAMENDO_NOTEBOOK_PAGE)
    providers.register('covers', COVERS_METHOD)


def disable(exaile):
    global JAMENDO_NOTEBOOK_PAGE, COVERS_METHOD
    providers.unregister('main-panel', JAMENDO_NOTEBOOK_PAGE)
    providers.unregister('covers', COVERS_METHOD)


class JamendoPanel(panel.Panel):

    __gsignals__ = {
        'append-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'download-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    ui_info = (os.path.dirname(__file__) + "/ui/jamendo_panel.ui", 'JamendoPanel')

    def __init__(self, parent, exaile):
        panel.Panel.__init__(self, parent, 'jamendo', "Jamendo")

        self.parent = parent
        self.exaile = exaile

        self.STATUS_READY = _("Ready")
        self.STATUS_SEARCHING = _("Searching Jamendo catalogue...")
        self.STATUS_RETRIEVING_DATA = _("Retrieving song data...")

        self.setup_widgets()

    # find out whats selected and add the tracks under it to the playlist
    def add_to_playlist(self):
        sel = self.get_selected_item()
        if isinstance(sel, jamtree.Artist):
            if not sel.expanded:
                self.expand_artist(sel, True)
                return

            for album in sel.albums:
                if not album.expanded:
                    self.expand_album(album, True)
                    return

            for album in sel.albums:
                track_list = []
                for track in album.tracks:
                    track_list.append(track)
                self.add_tracks_to_playlist(track_list)

        if isinstance(sel, jamtree.Album):
            if not sel.expanded:
                self.expand_album(sel, True)
                return
            track_list = []
            for track in sel.tracks:
                track_list.append(track)
            self.add_tracks_to_playlist(track_list)

        if isinstance(sel, jamtree.Track):
            self.add_track_to_playlist(sel)

    # is called when the user wants to download something
    def download_selected(self):
        print(
            'It would be really cool if this worked, unfortunately I still need to implement it.'
        )

    # initialise the widgets
    def setup_widgets(self):
        # connect to the signals we listen for
        self.builder.connect_signals(
            {
                'search_entry_activated': self.on_search_entry_activated,
                'search_entry_icon_release': self.clear_search_terms,
                'refresh_button_clicked': self.on_search_entry_activated,
                'search_combobox_changed': self.on_search_combobox_changed,
                'ordertype_combobox_changed': self.on_ordertype_combobox_changed,
                'orderdirection_combobox_changed': self.on_orderdirection_combobox_changed,
                'results_combobox_changed': self.on_results_combobox_changed,
            }
        )

        # set up the rightclick menu
        self.menu = menu.JamendoMenu(self)

        # setup images
        self.artist_image = icons.MANAGER.pixbuf_from_icon_name(
            'artist', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.album_image = icons.MANAGER.pixbuf_from_icon_name(
            'media-optical', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.title_image = icons.MANAGER.pixbuf_from_icon_name(
            'audio-x-generic', Gtk.IconSize.SMALL_TOOLBAR
        )

        # setup search combobox
        self.search_combobox = self.builder.get_object('searchComboBox')
        self.search_combobox.set_active(
            settings.get_option('plugin/jamendo/searchtype', 0)
        )

        # get handle on search entrybox
        self.search_textentry = self.builder.get_object('searchEntry')
        self.search_textentry.set_text(
            settings.get_option('plugin/jamendo/searchterms', "")
        )

        # setup order_by comboboxes
        self.orderby_type_combobox = self.builder.get_object('orderTypeComboBox')
        self.orderby_type_combobox.set_active(
            settings.get_option('plugin/jamendo/ordertype', 0)
        )
        self.orderby_direction_combobox = self.builder.get_object(
            'orderDirectionComboBox'
        )
        self.orderby_direction_combobox.set_active(
            settings.get_option('plugin/jamendo/orderdirection', 0)
        )

        # setup num_results combobox
        self.numresults_spinbutton = self.builder.get_object('numResultsSpinButton')
        self.numresults_spinbutton.set_value(
            settings.get_option('plugin/jamendo/numresults', 10)
        )

        # setup status label
        self.status_label = self.builder.get_object('statusLabel')
        self.set_status(self.STATUS_READY)

        # setup results treeview
        self.treeview = DragTreeView(self)
        self.treeview.connect("row-expanded", self.row_expanded)
        self.treeview.set_headers_visible(False)
        container = self.builder.get_object('treeview_box')
        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.add(self.treeview)
        scroll.set_shadow_type(Gtk.ShadowType.IN)
        container.pack_start(scroll, True, True, 0)
        container.show_all()

        selection = self.treeview.get_selection()
        selection.set_mode(Gtk.SelectionMode.SINGLE)
        pb = Gtk.CellRendererPixbuf()
        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn('Text')
        col.pack_start(pb, False)
        col.pack_start(cell, True)
        col.set_attributes(pb, pixbuf=0)
        col.set_attributes(cell, text=1)
        self.treeview.append_column(col)

        self.model = Gtk.TreeStore(GdkPixbuf.Pixbuf, str, GObject.TYPE_PYOBJECT)
        self.treeview.set_model(self.model)

    def set_status(self, message):
        self.status_label.set_text(message)

    def on_search_combobox_changed(self, box, params=None):
        settings.set_option('plugin/jamendo/searchtype', box.get_active())

    def on_ordertype_combobox_changed(self, box, params=None):
        settings.set_option('plugin/jamendo/ordertype', box.get_active())

    def on_orderdirection_combobox_changed(self, box, params=None):
        settings.set_option('plugin/jamendo/orderdirection', box.get_active())

    def on_results_combobox_changed(self, box, params=None):
        settings.set_option('plugin/jamendo/numresults', box.get_active())

    # is called whenever the user expands a row in the TreeView
    def row_expanded(self, tree, iter, path):
        sel = self.get_selected_item()
        if not sel.expanded:
            # unexpand node, will get expanded once contents are loaded
            self.expand_node(sel, False)
            if isinstance(sel, jamtree.Artist):
                self.expand_artist(sel, False)

            if isinstance(sel, jamtree.Album):
                self.expand_album(sel, False)

            if isinstance(sel, jamtree.Track):
                self.add_track_to_playlist(sel)

    # Expand an artist node (fetch albums for that artist)
    # artist: The jamtree.Artist object you want to expand the node for
    # add_to_playlist: Whether or not add_to_playlist() should be called when done
    def expand_artist(self, artist, add_to_playlist=False):
        self.set_status(self.STATUS_RETRIEVING_DATA)
        artist.expanded = True
        jamapi_thread = jamapi.get_albums(
            artist, self.expand_artist_callback, add_to_playlist
        )
        jamapi_thread.start()

    # Callback function for when the jamapi thread started in expand_artist() completes
    # artist: The jamtree.Artist object that should have had its albums populated by the jamapi thread
    def expand_artist_callback(self, artist, add_to_playlist=False):
        self.remove_dummy(artist)
        for album in artist.albums:
            parent = self.model.append(
                artist.row_pointer, (self.album_image, album.name, album)
            )
            album.row_pointer = parent
            self.model.append(parent, (self.title_image, "", ""))
        if add_to_playlist:
            self.add_to_playlist()
        self.expand_node(artist)
        self.set_status(self.STATUS_READY)

    # Expand an Album node (fetch tracks for album)
    # album: the Album object to get tracks for
    # add_to_playlist: Whether or not add_to_playlist() should be called when done
    def expand_album(self, album, add_to_playlist=False):
        self.set_status(self.STATUS_RETRIEVING_DATA)
        album.expanded = True
        jamapi_thread = jamapi.get_tracks(
            album, self.expand_album_callback, add_to_playlist
        )
        jamapi_thread.start()

    # Callback function for when the jamapi thread started in expand_album() completes
    # album: The jamtree.Album object that should have had its tracks populated by the jamapi thread
    def expand_album_callback(self, album, add_to_playlist=False):
        self.remove_dummy(album)
        for track in album.tracks:
            parent = self.model.append(
                album.row_pointer, (self.title_image, track.name, track)
            )
            track.row_pointer = parent
        if add_to_playlist:
            self.add_to_playlist()
        self.expand_node(album)
        self.set_status(self.STATUS_READY)

    # removes the first child node of a node
    def remove_dummy(self, node):
        iter = node.row_pointer
        dummy = self.model.iter_children(iter)
        self.model.remove(dummy)

    # expands a TreeView node
    def expand_node(self, node, expand=True):
        iter = node.row_pointer
        path = self.model.get_path(iter)
        if expand:
            self.treeview.expand_row(path, False)
        else:
            self.treeview.collapse_row(path)

    # is called when a user doubleclicks an item in the TreeView
    def button_press(self, widget, event):

        if event.type == Gdk.EventType._2BUTTON_PRESS:
            self.add_to_playlist()

    # is called by the search thread when it completed
    def response_callback(self, collection):
        self.set_status(self.STATUS_READY)

        if collection is None:
            return

        for item in collection:
            # add item to treeview
            image = self.artist_image

            if isinstance(item, jamtree.Album):
                image = self.album_image
            if isinstance(item, jamtree.Track):
                image = self.title_image

            parent = self.model.append(None, (image, item.name, item))
            item.row_pointer = parent

            if not isinstance(item, jamtree.Track):
                self.model.append(parent, (self.artist_image, "", ""))

    # retrieve and display search results
    def on_search_entry_activated(self, widget):
        self.set_status(self.STATUS_SEARCHING)

        # clear existing search
        self.model.clear()

        # get type of search
        iter = self.search_combobox.get_active_iter()
        search_type = self.search_combobox.get_model().get_value(iter, 0)
        iter = self.orderby_type_combobox.get_active_iter()
        orderby = self.orderby_type_combobox.get_model().get_value(iter, 0)
        iter = self.orderby_direction_combobox.get_active_iter()
        direction = self.orderby_direction_combobox.get_model().get_value(iter, 0)
        orderby += "_" + direction
        numresults = self.numresults_spinbutton.get_value_as_int()
        search_term = self.search_textentry.get_text()

        # save search term
        settings.set_option('plugin/jamendo/searchterms', search_term)

        if search_type == 'artist':
            resultthread = jamapi.get_artist_list(
                search_term, orderby, numresults, self.response_callback
            )
            resultthread.start()

        if search_type == 'album':
            resultthread = jamapi.get_album_list(
                search_term, orderby, numresults, self.response_callback
            )
            resultthread.start()

        if search_type == 'genre_tags':
            resultthread = jamapi.get_artist_list_by_genre(
                search_term, orderby, numresults, self.response_callback
            )
            resultthread.start()

        if search_type == 'track':
            resultthread = jamapi.get_track_list(
                search_term, orderby, numresults, self.response_callback
            )
            resultthread.start()

    # clear the search box and results
    def clear_search_terms(self, entry, icon_pos, event):
        entry.set_text('')

    # get the Object (Artist, Album, Track) associated with the currently
    # selected item in the TreeView
    def get_selected_item(self):
        iter = self.treeview.get_selection().get_selected()[1]
        return self.model.get_value(iter, 2)

    # get the path for the currently selected item in the TreeView
    def get_selected_item_path(self):
        iter = self.treeview.get_selection().get_selected()[1]
        return self.model.get_path(iter)

    # get the type of an object
    def typeof(self, something):
        return something.__class__

    # add a track to the playlist based on its url.
    # track: a jamtree.Track object
    def add_track_to_playlist(self, track):
        self.add_tracks_to_playlist([track])

    # add a bunch of tracks to the playlist at once
    # track_list: a python list of jamtree.Track objects
    def add_tracks_to_playlist(self, track_list):
        # convert list to list of xl.Track objects as opposed to jamtree.Track objects
        xltrack_list = []
        for track in track_list:
            tr = xltrack.Track(track.url, scan=False)
            tr.set_tags(
                title=track.name, artist=track.artist_name, album=track.album_name
            )
            xltrack_list.append(tr)
        self.exaile.gui.main.get_selected_page().playlist.extend(xltrack_list)

    # dragdrop stuff
    def drag_data_received(self, *e):
        pass

    def drag_data_delete(self, *e):
        pass

    def drag_get_data(self, treeview, context, selection, target_id, etime):
        self.add_to_playlist()


# The following is a custom CoverSearchMethod to retrieve covers from Jamendo
# It is designed to only to get covers for streaming tracks from jamendo


class JamendoCoverSearch(CoverSearchMethod):
    name = 'jamendo'
    use_cache = False  # do this since the tracks dont stay on local.
    fixed = True
    fixed_priority = 5  # take precendence, since we know we are 'right'
    # for matching tracks.

    def __init__(self, user_agent):
        self.user_agent = user_agent

    def find_covers(self, track, limit=-1):
        jamendo_url = track.get_loc_for_io()
        # http://stream10.jamendo.com/stream/61541/ogg2/02%20-%20PieRreF%20-%20Hologram.ogg?u=0&h=f2b227d38d
        split = jamendo_url.split('/')
        if len(split) > 5 and split[0] == 'http:' and split[2].endswith('.jamendo.com'):

            track_num = split[4]
            image_url = jamapi.get_album_image_url_from_track(track_num)

            if image_url:
                return [image_url]
        return []

    def get_cover_data(self, url):
        return common.get_url_contents(url, self.user_agent)



"""
A pure-python library to assist sending data to AudioScrobbler (the Last.fm
backend)
"""
import urllib.parse
import urllib.request
import logging
from time import mktime
from datetime import datetime, timedelta
from hashlib import md5

logger = logging.getLogger(__name__)

SESSION_ID = None
INITIAL_URL = None
POST_URL = None
POST_URL = None
NOW_URL = None
HARD_FAILS = 0
LAST_HS = None  # Last handshake time
HS_DELAY = 0  # wait this many seconds until next handshake
SUBMIT_CACHE = []
MAX_CACHE = 5  # keep only this many songs in the cache
MAX_SUBMIT = 10  # submit at most this many tracks at one time
PROTOCOL_VERSION = '1.2'
__LOGIN = {}  # data required to login

USER_AGENT_HEADERS = None


class BackendError(Exception):
    "Raised if the AS backend does something funny"
    pass


class AuthError(Exception):
    "Raised on authencitation errors"
    pass


class PostError(Exception):
    "Raised if something goes wrong when posting data to AS"
    pass


class SessionError(Exception):
    "Raised when problems with the session exist"
    pass


class ProtocolError(Exception):
    "Raised on general Protocol errors"
    pass


def set_user_agent(s):
    global USER_AGENT_HEADERS
    USER_AGENT_HEADERS = {'User-Agent': s}


def login(user, password, hashpw=False, client=('exa', '0.3.0'), post_url=None):
    """Authencitate with AS (The Handshake)

    @param user:     The username
    @param password: md5-hash of the user-password
    @param hashpw:   If True, then the md5-hash of the password is performed
                     internally. If set to False (the default), then the module
                     assumes the passed value is already a valid md5-hash of the
                     password.
    @param client:   Client information (see https://www.last.fm/api for more info)
    @type  client:   Tuple: (client-id, client-version)"""
    global LAST_HS, SESSION_ID, POST_URL, NOW_URL, HARD_FAILS, HS_DELAY, PROTOCOL_VERSION, INITIAL_URL, __LOGIN

    __LOGIN['u'] = user
    __LOGIN['c'] = client

    if LAST_HS is not None:
        next_allowed_hs = LAST_HS + timedelta(seconds=HS_DELAY)
        if datetime.now() < next_allowed_hs:
            delta = next_allowed_hs - datetime.now()
            raise ProtocolError(
                """Please wait another %d seconds until next handshake
(login) attempt."""
                % delta.seconds
            )

    LAST_HS = datetime.now()

    tstamp = int(mktime(datetime.now().timetuple()))
    # Store and keep first passed URL for future login retries
    INITIAL_URL = INITIAL_URL or post_url
    # Use passed or previously stored URL
    url = post_url or POST_URL

    if hashpw is True:
        __LOGIN['p'] = md5(password.encode('utf-8')).hexdigest()
    else:
        __LOGIN['p'] = password

    token = "%s%d" % (__LOGIN['p'], int(tstamp))
    token = md5(token.encode('utf-8')).hexdigest()
    values = {
        'hs': 'true',
        'p': PROTOCOL_VERSION,
        'c': client[0],
        'v': client[1],
        'u': user,
        't': tstamp,
        'a': token,
    }
    data = urllib.parse.urlencode(values)
    req = urllib.request.Request("%s?%s" % (url, data), None, USER_AGENT_HEADERS)
    response = urllib.request.urlopen(req)
    result = response.read().decode('utf-8')
    lines = result.split('\n')

    if lines[0] == 'BADAUTH':
        raise AuthError('Bad username/password')

    elif lines[0] == 'BANNED':
        raise Exception(
            '''This client-version was banned by Audioscrobbler. Please
contact the author of this module!'''
        )

    elif lines[0] == 'BADTIME':
        raise ValueError(
            '''Your system time is out of sync with Audioscrobbler.
Consider using an NTP-client to keep you system time in sync.'''
        )

    elif lines[0].startswith('FAILED'):
        handle_hard_error()
        raise BackendError("Authencitation with AS failed. Reason: %s" % lines[0])

    elif lines[0] == 'OK':
        # wooooooohooooooo. We made it!
        SESSION_ID = lines[1]
        NOW_URL = lines[2]
        POST_URL = lines[3]
        HARD_FAILS = 0
        logger.info("Logged in successfully to AudioScrobbler (%s)", url)

    else:
        # some hard error
        handle_hard_error()


def handle_hard_error():
    "Handles hard errors."
    global SESSION_ID, HARD_FAILS, HS_DELAY

    if HS_DELAY == 0:
        HS_DELAY = 60
    elif HS_DELAY < 120 * 60:
        HS_DELAY *= 2
    if HS_DELAY > 120 * 60:
        HS_DELAY = 120 * 60

    HARD_FAILS += 1
    if HARD_FAILS == 3:
        SESSION_ID = None


def now_playing(
    artist, track, album="", length="", trackno="", mbid="", inner_call=False
):
    """Tells audioscrobbler what is currently running in your player. This won't
    affect the user-profile on last.fm. To do submissions, use the "submit"
    method

    @param artist:  The artist name
    @param track:   The track name
    @param album:   The album name
    @param length:  The song length in seconds
    @param trackno: The track number
    @param mbid:    The MusicBrainz Track ID
    @return: True on success, False on failure"""

    global SESSION_ID, NOW_URL, INITIAL_URL

    if SESSION_ID is None:
        raise AuthError("Please 'login()' first. (No session available)")

    if POST_URL is None:
        raise PostError("Unable to post data. Post URL was empty!")

    if length != "" and not isinstance(length, type(1)):
        raise TypeError("length should be of type int")

    if trackno != "" and not isinstance(trackno, type(1)):
        raise TypeError("trackno should be of type int")

    # Quote from AS Protocol 1.1, Submitting Songs:
    #     Note that all the post variables noted here MUST be
    #     supplied for each entry, even if they are blank.
    track = track or ''
    artist = artist or ''
    album = album or ''

    values = {
        's': SESSION_ID,
        'a': artist,
        't': track,
        'b': album,
        'l': length,
        'n': trackno,
        'm': mbid,
    }
    data = urllib.parse.urlencode(values)

    req = urllib.request.Request(NOW_URL, data.encode('utf-8'), USER_AGENT_HEADERS)
    response = urllib.request.urlopen(req)
    result = response.read().decode('utf-8')

    if result.strip() == "OK":
        logger.info("Submitted \"Now Playing\" successfully to AudioScrobbler")
        return True
    elif result.strip() == "BADSESSION":
        if inner_call is False:
            login(__LOGIN['u'], __LOGIN['p'], client=__LOGIN['c'], post_url=INITIAL_URL)
            now_playing(artist, track, album, length, trackno, mbid, inner_call=True)
        else:
            raise SessionError('Invalid session')
    else:
        logger.warning("Error submitting \"Now Playing\"")

    return False


def submit(
    artist,
    track,
    time=0,
    source='P',
    rating="",
    length="",
    album="",
    trackno="",
    mbid="",
    autoflush=False,
):
    """Append a song to the submission cache. Use 'flush()' to send the cache to
    AS. You can also set "autoflush" to True.

    From the Audioscrobbler protocol docs:
    ---------------------------------------------------------------------------

    The client should monitor the user's interaction with the music playing
    service to whatever extent the service allows. In order to qualify for
    submission all of the following criteria must be met:

    1. The track must be submitted once it has finished playing. Whether it has
      finished playing naturally or has been manually stopped by the user is
      irrelevant.
    2. The track must have been played for a duration of at least 240 seconds or
      half the track's total length, whichever comes first. Skipping or pausing
      the track is irrelevant as long as the appropriate amount has been played.
    3. The total playback time for the track must be more than 30 seconds. Do
      not submit tracks shorter than this.
    4. Unless the client has been specially configured, it should not attempt to
      interpret filename information to obtain metadata instead of tags (ID3,
      etc).

    @param artist: Artist name
    @param track:  Track name
    @param time:   Time the track *started* playing in the UTC timezone (see
                  datetime.utcnow()).

                  Example: int(time.mktime(datetime.utcnow()))
    @param source: Source of the track. One of:
                  'P': Chosen by the user
                  'R': Non-personalised broadcast (e.g. Shoutcast, BBC Radio 1)
                  'E': Personalised recommendation except Last.fm (e.g.
                       Pandora, Launchcast)
                  'L': Last.fm (any mode). In this case, the 5-digit Last.fm
                       recommendation key must be appended to this source ID to
                       prove the validity of the submission (for example,
                       "L1b48a").
                  'U': Source unknown
    @param rating: The rating of the song. One of:
                  'L': Love (on any mode if the user has manually loved the
                       track)
                  'B': Ban (only if source=L)
                  'S': Skip (only if source=L)
                  '':  Not applicable
    @param length: The song length in seconds
    @param album:  The album name
    @param trackno:The track number
    @param mbid:   MusicBrainz Track ID
    @param autoflush: Automatically flush the cache to AS?
    @return:       True on success, False if something went wrong
    """
    if None in (artist, track):
        raise Exception

    if not artist.strip() or not track.strip():
        raise Exception

    global SUBMIT_CACHE, MAX_CACHE

    source = source.upper()
    rating = rating.upper()

    if source == 'L' and (rating == 'B' or rating == 'S'):
        raise ProtocolError(
            """You can only use rating 'B' or 'S' on source 'L'.
    See the docs!"""
        )

    if source == 'P' and length == '':
        raise ProtocolError(
            """Song length must be specified when using 'P' as
    source!"""
        )

    if not isinstance(time, type(1)):
        raise ValueError(
            """The time parameter must be of type int (unix
    timestamp). Instead it was %s"""
            % time
        )

    album = album or ''

    SUBMIT_CACHE.append(
        {
            'a': artist,
            't': track,
            'i': time,
            'o': source,
            'r': rating,
            'l': length,
            'b': album,
            'n': trackno,
            'm': mbid,
        }
    )

    if autoflush or len(SUBMIT_CACHE) >= MAX_CACHE:
        return flush()
    else:
        return True


def flush(inner_call=False):
    """Sends the cached songs to AS.

    @param inner_call: Internally used variable. Don't touch!"""
    global SUBMIT_CACHE, __LOGIN, MAX_SUBMIT, POST_URL, INITIAL_URL

    if POST_URL is None:
        raise ProtocolError(
            '''Cannot submit without having a valid post-URL. Did
you login?'''
        )

    values = {}

    for i, item in enumerate(SUBMIT_CACHE[:MAX_SUBMIT]):
        for key in item:
            values[key + "[%d]" % i] = item[key]

    values['s'] = SESSION_ID

    data = urllib.parse.urlencode(values)
    req = urllib.request.Request(POST_URL, data.encode('utf-8'), USER_AGENT_HEADERS)
    response = urllib.request.urlopen(req)
    result = response.read().decode('utf-8')
    lines = result.split('\n')

    if lines[0] == "OK":
        SUBMIT_CACHE = SUBMIT_CACHE[MAX_SUBMIT:]
        logger.info("AudioScrobbler OK: %s", data)
        return True
    elif lines[0] == "BADSESSION":
        if inner_call is False:
            login(__LOGIN['u'], __LOGIN['p'], client=__LOGIN['c'], post_url=INITIAL_URL)
            flush(inner_call=True)
        else:
            raise Warning("Infinite loop prevented")
    elif lines[0].startswith('FAILED'):
        handle_hard_error()
        raise BackendError("Submission to AS failed. Reason: %s" % lines[0])
    else:
        # some hard error
        handle_hard_error()
        return False


if __name__ == "__main__":
    login('user', 'password')
    submit('De/Vision', 'Scars', 1192374052, source='P', length=3 * 60 + 44)
    submit(
        'Spineshank',
        'Beginning of the End',
        1192374052 + (5 * 60),
        source='P',
        length=3 * 60 + 32,
    )
    submit(
        'Dry Cell',
        'Body Crumbles',
        1192374052 + (10 * 60),
        source='P',
        length=3 * 60 + 3,
    )
    print(flush())



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging
import pickle

from xl import common, event, playlist, settings

logger = logging.getLogger(__name__)


class PlayQueue(playlist.Playlist):
    """
    Manages the queue of songs to be played

    The content of the queue are processed before processing
    the content of the assigned playlist.

    When the remove_item_when_played option is enabled, the queue
    removes items from itself as they are played.

    When not enabled, the queue acts like a regular playlist, and
    moves the position as tracks are played.

    In this mode, when a new track is queued, the position is set
    to play that track, and play will continue with that track
    until the queue is exhausted, and then the assigned playlist
    will be continued.

    TODO: Queue needs to be threadsafe!
    """

    def __init__(self, player, name, location=None):
        playlist.Playlist.__init__(self, name=name)

        self.__queue_has_tracks_val = False
        self.__current_playlist = self  # this should never be None
        self.player = player

        # hack for making docs work
        if player is not None:
            player.queue = self

        if location is not None:
            self.load_from_location(location)

        event.add_callback(self._on_option_set, '%s_option_set' % name)

        self.__opt_remove_item_when_played = '%s/remove_item_when_played' % name
        self.__opt_disable_new_track_when_playing = (
            '%s/disable_new_track_when_playing' % name
        )
        self.__opt_enqueue_begins_playback = '%s/enqueue_begins_playback' % name

        self._on_option_set(None, settings, self.__opt_remove_item_when_played)
        self._on_option_set(None, settings, self.__opt_disable_new_track_when_playing)

    def _on_option_set(self, evtype, settings, option):
        if option == self.__opt_remove_item_when_played:
            self.__remove_item_on_playback = settings.get_option(option, True)
            if len(self):
                self.__queue_has_tracks = True
        elif option == self.__opt_disable_new_track_when_playing:
            self.__disable_new_track_when_playing = settings.get_option(option, False)

    def set_current_playlist(self, playlist):
        """
        Sets the playlist to be processed in the queue

        :param playlist: the playlist to process
        :type playlist: :class:`xl.playlist.Playlist`

        .. note:: The following :doc:`events </xl/event>` will be emitted by this method:

            * `queue_current_playlist_changed`: indicates that the queue playlist has been changed
        """
        if playlist is self.__current_playlist:
            return
        elif playlist is None:
            playlist = self

        if playlist is self:
            self.__queue_has_tracks = True

        self.__current_playlist = playlist
        event.log_event('queue_current_playlist_changed', self, playlist)

    #: The playlist currently processed in the queue
    current_playlist = property(
        lambda self: self.__current_playlist, set_current_playlist
    )

    def get_next(self):
        """
        Retrieves the next track that will be played. Does not
        actually set the position. When you call next(), it should
        return the same track.

        This exists to support retrieving a track before it actually
        needs to be played, such as for pre-buffering.

        :returns: the next track to be played
        :rtype: :class:`xl.trax.Track` or None
        """
        if self.__queue_has_tracks and len(self):
            if self.__remove_item_on_playback:
                return self[0]
            else:
                return playlist.Playlist.get_next(self)
        elif self.current_playlist is not self:
            return self.current_playlist.get_next()
        else:
            return None

    def next(self, autoplay=True, track=None):
        """
        Goes to the next track, either in the queue, or in the current
        playlist.  If a track is passed in, that track is played

        :param autoplay: play the track in addition to returning it
        :type autoplay: bool
        :param track: if passed, play this track
        :type track: :class:`xl.trax.Track`

        .. note:: The following :doc:`events </xl/event>` will be emitted by this method:

            * `playback_playlist_end`: indicates that the end of the queue has been reached
        """
        if track is None:
            if self.__queue_has_tracks:
                if not self.__remove_item_on_playback:
                    track = super().next()
                else:
                    try:
                        track = self.pop(0)
                    except IndexError:
                        pass

                # reached the end of the internal queue, don't repeat
                if track is None:
                    self.__queue_has_tracks = False

            if track is None and self.current_playlist is not self:
                track = self.current_playlist.next()

        if autoplay:
            self.player.play(track)

        if not track:
            event.log_event("playback_playlist_end", self, self.current_playlist)
        return track

    def prev(self):
        """
        Goes to the previous track
        """
        track = None
        if self.player.current:
            if self.player.get_time() < 5:
                if self.__queue_has_tracks and not self.__remove_item_on_playback:
                    position = self.current_position - 1
                    if position < 0:
                        position = 0 if len(self) else -1
                    self.current_position = position
                    track = self.current
                elif self.current_playlist is not self:
                    track = self.current_playlist.prev()

            if track is None:
                track = self.player.current
        else:
            track = self.current

        self.player.play(track)
        return track

    def get_current(self):
        """
        Gets the current track

        :returns: the current track
        :type: :class:`xl.trax.Track`
        """
        if self.player.current and self.current_position > 0:
            current = self.player.current
        else:
            current = playlist.Playlist.get_current(self)
            if current is None and self.current_playlist is not self:
                current = self.current_playlist.get_current()
        return current

    def get_current_position(self):
        if self.__remove_item_on_playback:
            return 0
        else:
            return playlist.Playlist.get_current_position(self)

    def set_current_position(self, position):
        if not self.__remove_item_on_playback:
            return playlist.Playlist.set_current_position(self, position)

    def is_play_enabled(self):
        ''':returns: True when calling play() will have no effect'''
        return not (self.player.is_playing() and self.__disable_new_track_when_playing)

    def play(self, track=None):
        """
        Starts queue processing with the given
        track preceding the queue content

        :param track: the track to play
        :type track: :class:`xl.trax.Track`
        """
        if self.player.is_playing():
            if not track or self.__disable_new_track_when_playing:
                return
        if not track:
            track = self.current
        if track:
            self.player.play(track)
            if self.__remove_item_on_playback:
                try:
                    del self[self.index(track)]
                except ValueError:
                    pass
        else:
            self.next()

    def queue_length(self):
        """
        Returns the number of tracks left to play in the queue's
        internal playlist.
        """
        if self.__remove_item_on_playback:
            return len(self)
        else:
            if not self.__queue_has_tracks:
                return -1
            else:
                return len(self) - (self.current_position + 1)

    def __set_queue_has_tracks(self, value):
        if value != self.__queue_has_tracks_val:
            oldpos = self.current_position
            self.__queue_has_tracks_val = value
            event.log_event(
                "playlist_current_position_changed",
                self,
                (self.current_position, oldpos),
            )

    # Internal value indicating whether the internal queue has tracks left to play
    __queue_has_tracks = property(
        lambda self: self.__queue_has_tracks_val, __set_queue_has_tracks
    )

    def __setitem__(self, i, value):
        """
        Overrides the playlist.Playlist list API.

        Allows us to ensure that when a track is added to an empty queue,
        we play it. Or not, depending on what the user wants.
        """
        old_len = playlist.Playlist.__len__(self)
        playlist.Playlist.__setitem__(self, i, value)

        # if nothing is queued, queue this track up
        if self.current_position == -1:
            if isinstance(i, slice):
                self.current_position = i.indices(len(self))[0] - 1
            else:
                self.current_position = i - 1

        self.__queue_has_tracks = True

        if (
            old_len == 0
            and settings.get_option('queue/enqueue_begins_playback', True)
            and old_len < playlist.Playlist.__len__(self)
        ):
            self.play()

    def _save_player_state(self, location):
        state = {}
        state['state'] = self.player.get_state()
        state['position'] = self.player.get_time()

        with open(location, 'wb') as f:
            pickle.dump(state, f, protocol=2)

    @common.threaded
    def _restore_player_state(self, location):
        if not settings.get_option("%s/resume_playback" % self.player._name, True):
            return

        try:
            with open(location, 'rb') as f:
                state = pickle.load(f)
        except Exception:
            return

        for req in ['state', 'position']:
            if req not in state:
                return

        self._do_restore_player_state(state)

    @common.idle_add()
    def _do_restore_player_state(self, state):

        if state['state'] in ['playing', 'paused']:

            start_at = None
            if state['position'] is not None:
                start_at = state['position']

            paused = state['state'] == 'paused' or settings.get_option(
                "%s/resume_paused" % self.player._name, False
            )

            self.player.play(self.get_current(), start_at=start_at, paused=paused)



# -*- coding: utf-8 -*-

from gi.repository import GdkPixbuf, GObject, Gtk, Gdk

from xlgui import icons


class CellRendererToggleImage(Gtk.CellRendererToggle):
    """
    Renders a toggleable state as an image
    """

    __gproperties__ = {
        'icon-name': (
            GObject.TYPE_STRING,
            'icon name',
            'The name of the themed icon to display. This '
            'property only has an effect if not overridden '
            'the "pixbuf" property.',
            '',
            GObject.ParamFlags.READWRITE,
        ),
        'pixbuf': (
            GdkPixbuf.Pixbuf,
            'pixbuf',
            'The pixbuf to render.',
            GObject.ParamFlags.READWRITE,
        ),
        'icon-size': (
            GObject.TYPE_UINT,
            'icon size',
            'The size of the rendered icon.',
            0,
            65535,
            Gtk.IconSize.SMALL_TOOLBAR,
            GObject.ParamFlags.READWRITE,
        ),
        'render-prelit': (
            GObject.TYPE_BOOLEAN,
            'render prelit',
            'Whether to render prelit states or not',
            True,
            GObject.ParamFlags.READWRITE,
        ),
    }

    def __init__(self):
        Gtk.CellRendererToggle.__init__(self)

        self.__icon_name = ''
        self.__pixbuf = None
        self.__insensitive_pixbuf = None
        self.__prelit_pixbuf = None
        self.__pixbuf_width = 0
        self.__pixbuf_height = 0
        self.__icon_size = Gtk.IconSize.SMALL_TOOLBAR
        self.__render_prelit = True

        # Any widget is fine
        self.__render_widget = Gtk.Button()
        self.__icon_theme = Gtk.IconTheme.get_default()

        self.set_property('activatable', True)

    def do_get_property(self, property):
        """
        Getter for custom properties
        """
        if property.name == 'icon-name':
            return self.__icon_name
        elif property.name == 'icon-size':
            return self.__icon_size
        elif property.name == 'pixbuf':
            return self.__pixbuf
        elif property.name == 'render-prelit':
            return self.__render_prelit
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_set_property(self, property, value):
        """
        Setter for custom properties
        """
        if property.name == 'icon-name':
            self.__icon_name = value
        elif property.name == 'icon-size':
            self.__icon_size = value
        elif property.name == 'pixbuf':
            self.__pixbuf = value
        elif property.name == 'render-prelit':
            self.__render_prelit = value
        else:
            raise AttributeError('unknown property %s' % property.name)

        self.__render_pixbufs()

    def __render_pixbufs(self):
        """
        Pre-renders all required pixbufs and caches them
        """
        # Get pixbuf from raw or name in that order
        if self.__pixbuf is None:
            pixbuf = icons.MANAGER.pixbuf_from_icon_name(
                self.__icon_name, self.__icon_size
            )
            if pixbuf is None:
                return
            self.__pixbuf = pixbuf

        # Cache dimensions
        self.__pixbuf_height = self.__pixbuf.get_height()
        self.__pixbuf_width = self.__pixbuf.get_width()

        self.__insensitive_pixbuf = self.__pixbuf.copy()
        # Render insensitive state
        self.__pixbuf.saturate_and_pixelate(
            dest=self.__insensitive_pixbuf, saturation=1, pixelate=True
        )

        if self.__render_prelit:
            self.__prelit_pixbuf = self.__pixbuf.copy()
            # Desaturate the active pixbuf
            self.__pixbuf.saturate_and_pixelate(
                dest=self.__prelit_pixbuf, saturation=0, pixelate=False
            )

    def do_render(self, cairo_context, widget, background_area, cell_area, flags):
        """
        Renders a custom toggle image
        """
        # Ensure pixbufs are rendered
        if self.__pixbuf is None:
            self.__render_pixbufs()

        pixbuf = None
        prelit = flags & Gtk.CellRendererState.PRELIT

        # Either draw the sensitive or insensitive state
        if self.props.sensitive:
            # Either draw the regular or the prelight state
            if self.props.active:
                pixbuf = self.__pixbuf
            elif self.__render_prelit and prelit:
                pixbuf = self.__prelit_pixbuf
        elif self.props.active:
            pixbuf = self.__insensitive_pixbuf

        # Do nothing if we are inactive or not prelit
        if pixbuf is not None:
            area_x, area_y, area_width, area_height = (
                cell_area.x,
                cell_area.y,
                cell_area.width,
                cell_area.height,
            )

            # Make sure to properly align the pixbuf
            x = area_x + area_width * self.props.xalign - self.__pixbuf_width // 2

            y = area_y + area_height * self.props.yalign - self.__pixbuf_height // 2

            Gdk.cairo_set_source_pixbuf(cairo_context, pixbuf, x, y)
            cairo_context.paint()



# Copyright (C) 2017 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import json
import time

from gi.repository import Gio

from xl.common import GioFileOutputStream
from xl.nls import gettext as _
from xl.trax import search

from xlgui.widgets import dialogs

from .gt_common import get_track_groups

import logging

logger = logging.getLogger(__name__)


def export_tags(exaile):
    """
    Exports all tags to a user specified JSON file
    """

    uri = dialogs.save(
        parent=exaile.gui.main.window,
        output_fname='tags.json',
        output_setting='plugin/grouptagger/export_dir',
        title=_('Export tags to JSON'),
        extensions={'.json': 'grouptagger JSON export'},
    )

    if uri is not None:

        # collect the data
        trackdata = {}
        for strack in search.search_tracks_from_string(exaile.collection, ''):
            track = strack.track
            tags = list(sorted(get_track_groups(track)))
            if tags:
                trackdata[track.get_loc_for_io()] = tags

        data = {
            '_meta': {
                'date': time.strftime('%Y-%m-%d %H:%M:%S'),
                'exporter': 'Exaile/%s' % exaile.get_version(),
                'version': 1,
            },
            'tracks': trackdata,
        }

        # save it
        with GioFileOutputStream(Gio.File.new_for_uri(uri), 'w') as fp:
            json.dump(data, fp, sort_keys=True, indent=4, separators=(',', ': '))

        logger.info("Exported tags of %s tracks to %s", len(trackdata), uri)



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import wave
import sunau
import os

from xl.metadata._base import BaseFormat, NotReadable

type_map = {"au": sunau, "wav": wave}


class WavFormat(BaseFormat):
    writable = False

    def load(self):
        ext = os.path.splitext(self.loc)[1][1:].lower()
        opener = type_map[ext]

        try:
            fp = open(self.loc, 'rb')
        except IOError:
            raise NotReadable

        try:
            with fp:
                f = opener.open(fp)
                length = f.getnframes() // f.getframerate()
            self.mutagen = {'__bitrate': -1, '__length': length}
        except (IOError, KeyError):
            self.mutagen = {'__bitrate': -1, '__length': -1}



try:
    from bs4 import BeautifulSoup
except ImportError:
    BeautifulSoup = None
import re
import urllib.parse

from xl.lyrics import LyricSearchMethod, LyricsNotFoundException
from xl import common, providers


def enable(exaile):
    """
    Enables the lyric wiki plugin that fetches track lyrics
    from lyrics.wikia.com
    """
    if BeautifulSoup:
        providers.register('lyrics', LyricWiki(exaile))
    else:
        raise NotImplementedError('BeautifulSoup is not available.')
        return False


def disable(exaile):
    providers.unregister('lyrics', providers.get_provider('lyrics', 'lyricwiki'))


class LyricWiki(LyricSearchMethod):

    name = "lyricwiki"
    display_name = "Lyric Wiki"

    def __init__(self, exaile):
        self.user_agent = exaile.get_user_agent_string('lyricwiki')

    def find_lyrics(self, track):
        try:
            (artist, title) = (
                track.get_tag_raw('artist')[0],
                track.get_tag_raw('title')[0],
            )
        except TypeError:
            raise LyricsNotFoundException

        if not artist or not title:
            raise LyricsNotFoundException

        artist = urllib.parse.quote(artist.replace(' ', '_'))
        title = urllib.parse.quote(title.replace(' ', '_'))

        url = 'https://lyrics.fandom.com/wiki/%s:%s' % (artist, title)

        try:
            source = common.get_url_contents(url, self.user_agent)
        except Exception:
            raise LyricsNotFoundException

        soup = BeautifulSoup(source, "lxml")
        lyrics = soup.findAll(attrs={"class": "lyricbox"})
        if lyrics:
            with_div = (
                lyrics[0].renderContents().decode('utf-8').replace('<br />', '\n')
            )
            string = '\n'.join(
                self.remove_div(with_div).replace('\n\n\n', '').split('\n')
            )
            lyrics = re.sub(r' Send.*?Ringtone to your Cell ', '', string)
        else:
            raise LyricsNotFoundException

        lyrics = self.remove_script(lyrics)
        lyrics = self.remove_html_tags(str(BeautifulSoup(lyrics, "lxml")))

        return (lyrics, self.name, url)



# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2013-2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import threading

from gi.repository import Gst
from gi.repository import GLib

import logging

logger = logging.getLogger(__name__)


class DynamicAudioSink(Gst.Bin):
    """
    An audio sink that can dynamically switch its output

    TODO: When switching outputs rapidly, sometimes it tends to seek
          ahead quite a bit. Not sure why.
    """

    def __init__(self, name):
        Gst.Bin.__init__(self, name=name)

        self.audio_sink = None
        self.__audio_sink_lock = threading.Lock()

        # Create an identity object so we don't need to deal with linking
        # the audio sink with anything external to this bin
        self.identity = Gst.ElementFactory.make('identity', None)
        self.identity.props.signal_handoffs = False
        self.add(self.identity)

        # Create a ghost sink pad so this bin appears to be an audio sink
        sinkpad = self.identity.get_static_pad("sink")
        self.add_pad(Gst.GhostPad.new('sink', sinkpad))

    def reconfigure(self, audio_sink):

        # don't try to switch more than one source at a time
        release_lock = True
        self.__audio_sink_lock.acquire()

        try:

            # If this is the first time we added a sink, just add it to
            # the pipeline and we're done.

            if not self.audio_sink:
                self._add_audiosink(audio_sink, None)
                return

            old_audio_sink = self.audio_sink

            # Ok, time to replace the old sink. If it's not in a playing state,
            # then this isn't so bad.

            # if we don't use the timeout, when we set it to READY, it may be performing
            # an async wait for PAUSE, so we use the timeout here.

            state = old_audio_sink.get_state(timeout=50 * Gst.MSECOND)[1]

            if state != Gst.State.PLAYING:
                buffer_position = None

                if state != Gst.State.NULL:
                    try:
                        buffer_position = old_audio_sink.query_position(Gst.Format.TIME)
                    except Exception:
                        pass

                self.remove(old_audio_sink)
                old_audio_sink.set_state(Gst.State.NULL)

                # Then add the new sink
                self._add_audiosink(audio_sink, buffer_position)

                return

            #
            # Otherwise, disconnecting the old device is a bit complex. Code is
            # derived from algorithm/code described at the following link:
            #
            # https://gstreamer.freedesktop.org/documentation/application-development/advanced/pipeline-manipulation.html
            #

            # Start off by blocking the src pad of the prior element
            spad = old_audio_sink.get_static_pad('sink').get_peer()
            spad.add_probe(
                Gst.PadProbeType.BLOCK_DOWNSTREAM, self._pad_blocked_cb, audio_sink
            )

            # Don't release the lock until pad block is done
            release_lock = False

        finally:
            if release_lock:
                self.__audio_sink_lock.release()

    def _pad_blocked_cb(self, pad, info, new_audio_sink):
        pad.remove_probe(info.id)

        old_audio_sink = self.audio_sink
        buffer_position = old_audio_sink.query_position(Gst.Format.TIME)

        # No data is flowing at this point. Unlink the element, add the new one
        self.remove(old_audio_sink)

        def _flush_old_sink():
            old_audio_sink.set_state(Gst.State.NULL)

        GLib.timeout_add(2000, _flush_old_sink)

        # Add the new element
        self._add_audiosink(new_audio_sink, buffer_position)

        self.__audio_sink_lock.release()

        # And drop the probe, which will cause data to flow again
        return Gst.PadProbeReturn.DROP

    def _add_audiosink(self, audio_sink, buffer_position):
        '''Sets up the new audiosink and syncs it'''

        self.add(audio_sink)
        audio_sink.sync_state_with_parent()
        self.identity.link(audio_sink)

        if buffer_position is not None:

            # buffer position is the output from get_position. If set, we
            # seek to that position.

            # TODO: this actually seems to skip ahead a tiny bit. why?

            # Note! this is super important in paused mode too, because when
            #       we switch the sinks around the new sink never goes into
            #       the paused state because there's no buffer. This forces
            #       a resync of the buffer, so things still work.

            seek_event = Gst.Event.new_seek(
                1.0,
                Gst.Format.TIME,
                Gst.SeekFlags.FLUSH,
                Gst.SeekType.SET,
                buffer_position[1],
                Gst.SeekType.NONE,
                0,
            )

            self.send_event(seek_event)

        self.audio_sink = audio_sink


# vim: et sts=4 sw=4
